__all__ = ('ApplicationCommandOption',)

from scarletio import RichAttributeErrorBaseType

from ..channel import ChannelType
from ..localization.helpers import get_localized_length, localized_dictionary_builder
from ..localization.utils import build_locale_dictionary, destroy_locale_dictionary, hash_locale_dictionary
from ..preconverters import preconvert_preinstanced_type

from .application_command_option_choice import ApplicationCommandOptionChoice
from .constants import (
    APPLICATION_COMMAND_CHOICES_MAX, APPLICATION_COMMAND_DESCRIPTION_LENGTH_MAX,
    APPLICATION_COMMAND_DESCRIPTION_LENGTH_MIN, APPLICATION_COMMAND_NAME_LENGTH_MAX,
    APPLICATION_COMMAND_NAME_LENGTH_MIN, APPLICATION_COMMAND_OPTIONS_MAX, APPLICATION_COMMAND_OPTION_MAX_LENGTH_MAX,
    APPLICATION_COMMAND_OPTION_MAX_LENGTH_MIN, APPLICATION_COMMAND_OPTION_MIN_LENGTH_MAX,
    APPLICATION_COMMAND_OPTION_MIN_LENGTH_MIN
)
from .helpers import apply_translation_into
from .preinstanced import ApplicationCommandOptionType


def _validate_max_length(max_length, type_):
    """
    Validates the given ``ApplicationCommandOption`` option's `max_length` value.
    
    Parameters
    ----------
    max_length : `None`, `int`
        The maximum input length allowed for this option.
        
        Only applicable for string options.
    
    type_ : ``ApplicationCommandOptionType``
        Respective application command option type.
    
    Returns
    -------
    max_length : `int`
    
    Raises
    ------
    TypeError
        - If `max_length`'s type is incorrect.
    ValueError
        - If `max_length` is not applicable for the given type.
    """
    if (max_length is None):
        return 0
    
    if not isinstance(max_length, int):
        raise TypeError(
            f'`max_length` can be `None`, `int`, got {max_length.__class__.__name__}; {max_length!r}.'
        )
        
    if max_length == 0:
        return 0
    
    if type_ is not ApplicationCommandOptionType.string:
        raise ValueError(
            f'`max_length` is only meaningful if `type` is {ApplicationCommandOptionType.string!r}, got '
            f'type_={type_!r}; max_length = {max_length!r}.'
        )
    
    if max_length < APPLICATION_COMMAND_OPTION_MAX_LENGTH_MIN:
        max_length = APPLICATION_COMMAND_OPTION_MAX_LENGTH_MIN
    
    elif max_length >= APPLICATION_COMMAND_OPTION_MAX_LENGTH_MAX:
        max_length = 0

    return max_length


def _validate_min_length(min_length, type_):
    """
    Validates the given ``ApplicationCommandOption`` option's `min_length` value.
    
    Parameters
    ----------
    min_length : `None`, `int`
        The minimum input length allowed for this option.
        
        Only applicable for string options.
    
    type_ : ``ApplicationCommandOptionType``
        Respective application command option type.
    
    Returns
    -------
    min_length : `int`
    
    Raises
    ------
    TypeError
        - If `min_length`'s type is incorrect.
    ValueError
        - If `min_length` is not applicable for the given type.
    """
    if (min_length is None):
        return 0
    
    if not isinstance(min_length, int):
        raise TypeError(
            f'`min_length` can be `None`, `int`, got {min_length.__class__.__name__}; {min_length!r}.'
        )
        
    if min_length == 0:
        return 0
    
    if type_ is not ApplicationCommandOptionType.string:
        raise ValueError(
            f'`min_length` is only meaningful if `type` is {ApplicationCommandOptionType.string!r}, got '
            f'type_={type_!r}; {min_length!r}.'
        )
    
    if min_length < APPLICATION_COMMAND_OPTION_MIN_LENGTH_MIN:
        min_length = APPLICATION_COMMAND_OPTION_MIN_LENGTH_MIN
    
    elif min_length > APPLICATION_COMMAND_OPTION_MIN_LENGTH_MAX:
        min_length = APPLICATION_COMMAND_OPTION_MIN_LENGTH_MAX

    return min_length


def _assert__application_command_option__autocomplete(autocomplete):
    """
    Asserts the `autocomplete` parameter of ``ApplicationCommandOption.__new__`` method.
    
    Parameters
    ----------
    autocomplete : `bool`
        Whether the application command option is autocompleted.
    
    Raises
    ------
    AssertionError
        - If `autocomplete` is not `bool`.
    """
    if not isinstance(autocomplete, bool):
        raise AssertionError(
            f'`autocomplete` can be `bool`, got {autocomplete.__class__.__name__}; {autocomplete!r}.'
        )
    
    return True


def _assert__application_command_option__default(default):
    """
    Asserts the `default` parameter of ``ApplicationCommandOption.__new__`` method.
    
    Parameters
    ----------
    default : `bool`
        Whether the application command option is the default one in it's group.
    
    Raises
    ------
    AssertionError
        - If `default` is not `bool`.
    """
    if not isinstance(default, bool):
        raise AssertionError(
            f'`default` can be `bool`, got {default.__class__.__name__}; {default!r}.'
        )
    
    return True


def _assert__application_command_option__description(description):
    """
    Asserts the `description` parameter of ``ApplicationCommandOption.__new__`` method.
    
    Parameters
    ----------
    description : `str`
        The application command option's description.
    
    Raises
    ------
    AssertionError
        - If `description` is not `str`.
        - if `description`'s length is out of the expected range.
    """
    if not isinstance(description, str):
        raise AssertionError(
            f'`description` can be `str`, got {description.__class__.__name__}; {description!r}.'
        )
    
    description_length = len(description)
    if (
        description_length < APPLICATION_COMMAND_DESCRIPTION_LENGTH_MIN or
        description_length > APPLICATION_COMMAND_DESCRIPTION_LENGTH_MAX
    ):
        raise AssertionError(
            f'`description` length can be in range '
            f'[{APPLICATION_COMMAND_DESCRIPTION_LENGTH_MIN}:{APPLICATION_COMMAND_DESCRIPTION_LENGTH_MAX}], '
            f'got {description_length!r}; {description!r}.'
        )

    return True


def _assert__application_command_option__name(name):
    """
    Asserts the `name` parameter of ``ApplicationCommandOption.__new__`` method.
    
    Parameters
    ----------
    name : `str`
        The application command option's name.
    
    Raises
    ------
    AssertionError
        - If `name` is not `str`.
        - if `name`'s length is out of the expected range.
    """
    if not isinstance(name, str):
        raise AssertionError(
            f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
        )
    
    name_length = len(name)
    if (
        name_length < APPLICATION_COMMAND_NAME_LENGTH_MIN or
        name_length > APPLICATION_COMMAND_NAME_LENGTH_MAX
    ):
        raise AssertionError(
            f'`name` length can be in range '
            f'[{APPLICATION_COMMAND_NAME_LENGTH_MIN}:{APPLICATION_COMMAND_NAME_LENGTH_MAX}], got '
            f'{name_length!r}; {name!r}.'
        )
    
    return True


def _assert__application_command_option__required(required):
    """
    Asserts the `required` parameter of ``ApplicationCommandOption.__new__`` method.
    
    Parameters
    ----------
    required : `bool`
        Whether the application command option is required.
    
    Raises
    ------
    AssertionError
        - If `required` is not `bool`.
    """
    if not isinstance(required, bool):
        raise AssertionError(
            f'`required` can be `bool`, got {required.__class__.__name__}; {required!r}.'
        )
    
    return True


def _assert__application_command_option__autocomplete__applicability(autocomplete, choices, type_):
    """
    Asserts whether the `autocomplete` parameter of the of ``ApplicationCommandOption.__new__`` method
    is applicable for the given `choices` and `type_` combinations.
    
    Parameters
    ----------
    autocomplete : `bool`
        Whether the option supports auto completion.
    choices : `None`, `list` of ``ApplicationCommandOptionChoice``
        The choices of the command for string or integer types.
    type_ : ``ApplicationCommandOptionType``
        The application command option's type.

    Raises
    ------
    AssertionError
        - If both `autocomplete` and `choices` are defined.
        - If `autocomplete` is defined, but the parameters' type is not string.
    """
    if autocomplete:
        if (choices is not None):
            raise AssertionError(
                f'`autocomplete` and `choices` parameters are mutually exclusive, got '
                f'autocomplete={autocomplete!r}; choices={choices!r}.'
            )
        
        if (type_ is not ApplicationCommandOptionType.string):
            raise AssertionError(
                f'`autocomplete` is only available for string option type, got type={type_!r}.'
            )
    
    return True


def _assert__application_command_option__channel_types__applicability(channel_types, type_):
    """
    Asserts whether the `autocomplete` parameter of the of ``ApplicationCommandOption.__new__`` method
    is applicable for the given `choices` and `type_` combinations.
    
    Parameters
    ----------
    channel_types : `None`, `tuple` of `int`
        The accepted channel types by the option.
    type_ : ``ApplicationCommandOptionType``
        The application command option's type.

    Raises
    ------
    AssertionError
        - If `channel_types` is given, but `type_` is not `ApplicationCommandOptionType.channel`.
    """
    if (channel_types is not None) and (type_ is not ApplicationCommandOptionType.channel):
        raise AssertionError(
            f'`channel_types` is only meaningful if `type_` is `{ApplicationCommandOptionType.__name__}.channel`, got '
            f'type_={type_!r}; channel_types = {channel_types!r}.'
        )

    return True


class ApplicationCommandOption(RichAttributeErrorBaseType):
    """
    An option of an ``ApplicationCommand``.
    
    Attributes
    ----------
    autocomplete : `bool`
        Whether the option supports auto completion.
        
        Mutually exclusive with the ``.choices``. Only applicable for string type parameters.
    
    channel_types : `None`, `tuple` of ``ChannelType``
        The accepted channel types by the option.
        
        Only applicable if ``.type`` is set to `ApplicationCommandOptionType.channel`.
    
    choices : `None`, `list` of ``ApplicationCommandOptionChoice``
        Choices for `str` and `int` types for the user to pick from.
        
        Mutually exclusive with the ``.autocomplete``.
    
    default : `bool`
        Whether the option is the default one. Only one option can be `default`.
    
    description : `str`
        The description of the application command option. It's length can be in range [1:100].
    
    description_localizations : `None`, `dict` of (``Locale``, `str`) items
        Localized descriptions of the option.
    
    max_length : `int`
        The maximum input length allowed for this option.
        
        Only applicable for string options.
    
    max_value : `None`, `int`, `float`
        The maximal value permitted for this option.
        
        Only applicable for integer as `int`, and for float options as `float`.
    
    min_length : `int`
        The minimum input length allowed for this option.
        
        Only applicable for string options.
    
    min_value : `None`, `int`, `float`
        The minimum value permitted for this option.
        
        Only applicable for integer as `int`, and for float options as `float`.
    
    name : `str`
        The name of the application command option. It's length can be in range [1:32].
    
    name_localizations : `None`, `dict` of (``Locale``, `str`) items
        Localized names of the option.
    
    options : `None`, `list` of ``ApplicationCommandOption``
        If the command's type is sub-command group type, then this nested option will be the parameters of the
        sub-command. It's length can be in range [0:25]. If would be set as empty list, instead is set as `None`.
    
    required : `bool`
        Whether the parameter is required. Defaults to `False`.
    
    type : ``ApplicationCommandOptionType``
        The option's type.
    """
    __slots__ = (
        'autocomplete', 'channel_types', 'choices', 'default', 'description', 'description_localizations',
        'max_length', 'max_value', 'min_length', 'min_value', 'name', 'name_localizations', 'options', 'required',
        'type'
    )
    
    def __new__(
        cls, name, description, type_, *, autocomplete=False, channel_types = None, choices=None, default = False,
        description_localizations=None, max_length = None, max_value = None, min_length = None, min_value = None,
        name_localizations=None, options=None, required=False
    ):
        """
        Creates a new ``ApplicationCommandOption`` with the given parameters.
        
        Parameters
        ----------
        name : `str`
            The name of the command. It's length can be in range [1:32].
        
        description : `str`
            The command's description. It's length can be in range [2:100].
        
        type_ : `int`, ``ApplicationCommandOptionType``
            The application command option's type.
        
        autocomplete : `bool` = `False`, Optional (Keyword only)
            Whether the option supports auto completion.
            
            Mutually exclusive with the `choices` parameter. Only applicable for string type parameters.
        
        channel_types : `None`, `iterable` of (``ChannelType``, `int`) = `None`, Optional (Keyword only)
            The accepted channel types by the option.
            
            Only applicable if ``.type`` is set to `ApplicationCommandOptionType.channel`.
        
        choices : `None`, `iterable` of ``ApplicationCommandOptionChoice`` = `None`, Optional (Keyword only)
            The choices of the command for string or integer types. It's length can be in range [0:25].
            
            Mutually exclusive with the `autocomplete` parameter.
        
        default : `bool` = `False`, Optional (Keyword only)
            Whether the option is the default one.
        
        description_localizations : `None`, `dict` of ((`str`, ``Locale``), `str`) items,
                (`list`, `set`, `tuple`) of `tuple` ((`str`, ``Locale``), `str`) = `None`, Optional (Keyword only)
            Localized descriptions of the option.
            
        max_length : `None`, `int` = `None`, Optional (Keyword only)
            The maximum input length allowed for this option.
            
            Only applicable for string options.
        
        max_value : `None`, `int`, `float` = `None`, Optional (Keyword only)
            The maximal value permitted for this option.
            
            Only applicable for integer as `int`, and for float options as `float`.
        
        min_length : `None`, `int` = `None`, Optional (Keyword only)
            The minimum input length allowed for this option.
            
            Only applicable for string options.
            
        min_value : `None`, `int`, `float` = `None`, Optional (Keyword only)
            The minimum value permitted for this option.
            
            Only applicable for integer as `int`, and for float options as `float`.
        
        name_localizations : `None`, `dict` of ((`str`, ``Locale``), `str`) items,
                (`list`, `set`, `tuple`) of `tuple` ((`str`, ``Locale``), `str`) = `None`, Optional (Keyword only)
            Localized names of the option.
        
        options : `None`, `iterable` of ``ApplicationCommandOption`` = `None`, Optional (Keyword only)
            The parameters of the command. It's length can be in range [0:25]. Only applicable for sub command groups.
        
        required : `bool` = `False`, Optional (Keyword only)
            Whether the parameter is required.
        
        Raises
        ------
        TypeError
            - If a parameter's type is unexpected.
            - If `options` was given meanwhile `type_` is not a sub-command group option type.
            - If a choice's value's type not matched the expected type described `type_`.
        ValueError
            - If `type_` was given as `int`, but it do not matches any of the precreated
                ``ApplicationCommandOptionType``-s.
            - If `channel_types` contains an unknown channel type value.
            - If `max_value` or `min_value` is given, but it is not applicable for the given `type_`.
            - If `name_localizations` has an item with incorrect structure.
            - If `description_localizations` has an item with incorrect structure.
            - If `max_length` or `min_length` is given, but it is not applicable for the given `type_`.
        """
        # autocomplete
        assert _assert__application_command_option__autocomplete(autocomplete)
        
        # channel_types
        channel_types_processed = None
        
        if (channel_types is not None):
            channel_types_processed = None
            
            if getattr(type(channel_types), '__iter__', None) is None:
                raise TypeError(
                    f'`channel_types` can be `None`, `iterable`, got '
                    f'{channel_types.__class__.__anme__}; {channel_types!r}.'
                )
            
            for channel_type in channel_types:
                if isinstance(channel_type, ChannelType):
                    pass
                
                elif isinstance(channel_type, ChannelType.VALUE_TYPE):
                    channel_type = ChannelType.get(channel_type)
                
                else:
                    raise TypeError(
                        f'`channel_types` can have `{ChannelType.__name__}`, `int` elements, '
                        f'got {channel_type.__class__.__name__}; {channel_type!r}; channel_types = {channel_types!r}.'
                    )
                
                if channel_types_processed is None:
                    channel_types_processed = set()
                
                channel_types_processed.add(channel_type)
            
            if (channel_types_processed is not None):
                channel_types_processed = tuple(sorted(channel_types_processed))
        
        
        # choices
        choices_processed = None
        
        if (choices is not None):
            if (getattr(choices, '__iter__', None) is None):
                raise TypeError(
                    f'`choices` can be `iterable of `{ApplicationCommandOptionChoice.__name__}`, '
                    f'got {choices.__class__.__name__}; {choices!r}.'
                )
            
            for choice in choices:
                if not isinstance(choice, ApplicationCommandOptionChoice):
                    raise TypeError(
                        f'`choices` can contain `{ApplicationCommandOptionChoice.__name__}` elements, got '
                        f'{choice.__class__.__name__}; {choice!r}; choices={choices!r}.'
                    )
                
                if choices_processed is None:
                    choices_processed = []
                
                choices_processed.append(choice)
            
            if (choices_processed is not None) and (len(choices_processed) > APPLICATION_COMMAND_CHOICES_MAX):
                del choices_processed[APPLICATION_COMMAND_CHOICES_MAX:]
        
        
        # default
        assert _assert__application_command_option__default(default)
        
        # description
        assert _assert__application_command_option__description(description)
        
        # description_localizations
        description_localizations = localized_dictionary_builder(description_localizations, 'description_localizations')
        
        # max_length
        # requires `type`
        
        # max_value
        # requires `type`
        
        # min_length
        # requires `type`
        
        # min_value
        # requires `type`
        
        # name
        assert _assert__application_command_option__name(name)
        
        # name_localizations
        name_localizations = localized_dictionary_builder(name_localizations, 'name_localizations')
        
        # options
        options_processed = None
        
        if (options is not None):
            if getattr(options, '__iter__', None) is None:
                raise TypeError(
                    f'`options` can be `None`, `iterable` of `{ApplicationCommandOption.__name__}`, '
                    f'got {options.__class__.__name__}; {options!r}.'
                )
            
            for option in options:
                if not isinstance(option, ApplicationCommandOption):
                    raise TypeError(
                        f'`options` contains a non `{ApplicationCommandOption.__name__}` element, got '
                        f'{option.__class__.__name__}; {option!r}; options={options!r}.'
                    )
                
                if (options_processed is None):
                    options_processed = []
                
                options_processed.append(option)
            
            if (options_processed is not None) and (len(options_processed) > APPLICATION_COMMAND_OPTIONS_MAX):
                # Deleting the excess should be fine.
                del options_processed[APPLICATION_COMMAND_OPTIONS_MAX:]
        
        
        # required
        assert _assert__application_command_option__required(required)
        
        # type
        type_ = preconvert_preinstanced_type(type_, 'type_', ApplicationCommandOptionType)
        
        # Postprocessing
        
        # max_length
        max_length = _validate_max_length(max_length, type_)
        
        # max_value
        if (max_value is not None):
            if type_ is ApplicationCommandOptionType.integer:
                if not isinstance(max_value, int):
                    raise TypeError(
                        f'`max_value` can be `int`, if `type_` is defined as {type_!r}, got '
                        f'{max_value.__class__.__name__}; {max_value!r}.'
                    )
            
            elif type_ is ApplicationCommandOptionType.float:
                if not isinstance(max_value, float):
                    raise TypeError(
                        f'`max_value` can be `float`, if `type_` is defined as {type_!r}, got '
                        f'{max_value.__class__.__name__}; {max_value!r}.'
                    )
            
            else:
                raise ValueError(
                    f'`max_value` is only meaningful if `type` is either '
                    f'{ApplicationCommandOptionType.integer!r}, or {ApplicationCommandOptionType.float!r}, got '
                    f'type_={type_!r}; max_value = {max_value!r}.'
                )
        
        # min_length
        min_length = _validate_min_length(min_length, type_)
        
        # min_value
        if (min_value is not None):
            if type_ is ApplicationCommandOptionType.integer:
                if not isinstance(min_value, int):
                    raise TypeError(
                        f'`min_value` can be `int` type, if `type_` is defined as {type_!r}, got '
                        f'{min_value.__class__.__name__}; {min_value!r}.'
                    )
            
            elif type_ is ApplicationCommandOptionType.float:
                if not isinstance(min_value, float):
                    raise TypeError(
                        f'`min_value` can be `float` type, if `type_` is defined as {type_!r}, got '
                        f'{min_value.__class__.__name__}; {min_value!r}.'
                    )
            
            else:
                raise ValueError(
                    f'`min_value` is only meaningful if `type` is either '
                    f'{ApplicationCommandOptionType.integer!r}, or {ApplicationCommandOptionType.float!r}, got '
                    f'type_={type_!r}; min_value = {min_value!r}.'
                )
        
        # postprocessing | autocomplete
        assert _assert__application_command_option__autocomplete__applicability(autocomplete, choices_processed, type_)

        
        # postprocessing | choices
        if (choices_processed is not None):
            if type_ is ApplicationCommandOptionType.string:
                expected_choice_type = str
            elif type_ is ApplicationCommandOptionType.integer:
                expected_choice_type = int
            elif type_ is ApplicationCommandOptionType.float:
                expected_choice_type = float
            else:
                raise TypeError(
                    f'`choices` can be bound either to string, integer or float option types, got '
                    f'choices={choices!r}, type={type_!r}.'
                )
            
            for index, choice in enumerate(choices):
                if not isinstance(choice.value, expected_choice_type):
                    raise TypeError(
                        f'`choices[{index!r}]` is not `{expected_choice_type.__name__}` as expected from the received'
                        f'command option type, got {choice.__class__.__name__}; {choice!r}; type_={type_!r}; '
                        f'choices={choices!r}.'
                    )
        
        # postprocessing | channel_types
        assert _assert__application_command_option__channel_types__applicability(channel_types_processed, type_)
        
        
        self = object.__new__(cls)
        
        self.autocomplete = autocomplete
        self.channel_types = channel_types_processed
        self.choices = choices_processed
        self.default = default
        self.description = description
        self.description_localizations = description_localizations
        self.max_length = max_length
        self.max_value = max_value
        self.min_length = min_length
        self.min_value = min_value
        self.name = name
        self.name_localizations = name_localizations
        self.options = options_processed
        self.required = required
        self.type = type_
        
        return self
    
    
    def add_option(self, option):
        """
        Adds a new option to the application command option.
        
        Parameters
        ----------
        option : ``ApplicationCommandOption``
            The option to add.
        
        Returns
        -------
        self : ``ApplicationCommandOption``
        
        Raises
        ------
        TypeError
            - If the source application command's type is not a sub-command group type.
            - If `option` is not ``ApplicationCommandOption``.
            - If `option` is a sub command group option.
        """
        if self.type is not ApplicationCommandOptionType.sub_command_group:
            raise TypeError(
                f'`option` can be added only if the command option\s type is sub command option, '
                f'got option={option!r}, self={self!r}.'
            )
        
        if not isinstance(option, ApplicationCommandOption):
            raise TypeError(
                f'`option` can be `{ApplicationCommandOption.__name__}`, got '
                f'{option.__class__.__name__}; {option!r}.'
            )
    
        if option.type is ApplicationCommandOptionType.sub_command_group:
            raise TypeError(
                f'`option`\'s type is sub-command group option, but sub-command groups cannot be '
                f'added under sub-command groups; got {option!r}; self={self!r}.'
            )
        
        options = self.options
        if options is None:
            options = []
            self.options = options
        
        if len(options) < APPLICATION_COMMAND_OPTIONS_MAX:
            options.append(option)
        
        return self
    
    
    def add_choice(self, choice):
        """
        Adds a ``ApplicationCommandOptionChoice`` to the application command option.
        
        Parameters
        ----------
        choice : ``ApplicationCommandOptionChoice``, `tuple` (`str`, (`str`, `int`, `float`))
            The choice to add.
        
        Returns
        -------
        self : ``ApplicationCommandOption``
        
        Raises
        ------
        TypeError
            - If the source application command's type is not a string, int nor float group type.
            - If the `choice`'s value's type is not the expected one by the command option's type.
            - If `choice`'s type is neither ``ApplicationCommandOptionChoice`` nor a `tuple` representing it's `.name`
                and `.value`.
        """
        if isinstance(choice, ApplicationCommandOptionChoice):
            pass
        
        elif isinstance(choice, tuple):
            if len(choice) != 2:
                raise TypeError(
                    f'`choice` `tuple`\'s length should be `2` representing a '
                    f'`{ApplicationCommandOptionChoice.__name__}`\'s `.name` and `.value`, got {choice!r}.'
                )
            
            choice = ApplicationCommandOptionChoice(*choice)
        else:
            raise TypeError(
                f'`choice` can be `{ApplicationCommandOptionChoice.__name__}`, '
                f'`tuple` (`str`, (`str`, `int`, `float`)), got '
                f'{choice.__class__.__name__}; {choice!r}.'
            )
        
        type_ = self.type
        if type_ is ApplicationCommandOptionType.string:
            expected_choice_type = str
        elif type_ is ApplicationCommandOptionType.integer:
            expected_choice_type = int
        elif type_ is ApplicationCommandOptionType.float:
            expected_choice_type = float
        else:
            raise TypeError(
                f'`choice` is bound to string, integer and float application option types, '
                f'got choice={choice!r}, self={self!r}.'
            )
        
        if not isinstance(choice.value, expected_choice_type):
            raise TypeError(
                f'`choice` value\'s type is not `{expected_choice_type.__name__}` as expected, got '
                f'choice={choice!r}, self={self!r}.'
            )
        
        choices = self.choices
        if choices is None:
            choices = []
            self.choices = choices
        
        if len(choices) < APPLICATION_COMMAND_CHOICES_MAX:
            choices.append(choice)
        
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``ApplicationCommandOption`` from the received data from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received application command option data.
        
        Returns
        -------
        self : ``ApplicationCommandOption``
            The created application command option.
        """
        # autocomplete
        autocomplete = data.get('autocomplete', False)
        
        # channel_types
        channel_types = data.get('channel_types', None)
        if (channel_types is None) or (not channel_types):
            channel_types = None
        else:
            channel_types = tuple(sorted(ChannelType.get(channel_type) for channel_type in channel_types))
        
        # choices
        choice_datas = data.get('choices', None)
        if (choice_datas is None) or (not choice_datas):
            choices = None
        else:
            choices = [ApplicationCommandOptionChoice.from_data(choice_data) for choice_data in choice_datas]
        
        # default
        default = data.get('default', False)
        
        # description
        description = data['description']
        
        # description_localizations
        description_localizations = build_locale_dictionary(data.get('description_localizations', None))
        
        # max_length
        max_length = data.get('max_length', None)
        if (max_length is None) or (max_length == APPLICATION_COMMAND_OPTION_MAX_LENGTH_MAX):
            max_length = 0
        
        # max_value
        max_value = data.get('max_value', None)
        
        # min_length
        min_length = data.get('min_length', None)
        if (min_length is None):
            min_length = 0
        
        # min_value
        min_value = data.get('min_value', None)
        
        # name
        name = data['name']
        
        # name_localizations
        name_localizations = build_locale_dictionary(data.get('name_localizations', None))
        
        # options
        option_datas = data.get('options', None)
        if (option_datas is None) or (not option_datas):
            options = None
        else:
            options = [ApplicationCommandOption.from_data(option_data) for option_data in option_datas]
        
        # required
        required = data.get('required', False)
        
        # type
        type_ = ApplicationCommandOptionType.get(data['type'])
        
        
        self = object.__new__(cls)
        
        self.autocomplete = autocomplete
        self.channel_types = channel_types
        self.choices = choices
        self.default = default
        self.description = description
        self.description_localizations = description_localizations
        self.max_length = max_length
        self.max_value = max_value
        self.min_length = min_length
        self.min_value = min_value
        self.name = name
        self.name_localizations = name_localizations
        self.options = options
        self.required = required
        self.type = type_
        
        return self
    
    
    def to_data(self):
        """
        Converts the application command option to a json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        data = {}
        
        # autocomplete
        if self.autocomplete:
            data['autocomplete'] = True
        
        # channel_types
        channel_types = self.channel_types
        if (channel_types is not None):
            data['channel_types'] = [channel_type.value for channel_type in channel_types]
        
        # choices
        choices = self.choices
        if (choices is not None):
            data['choices'] = [choice.to_data() for choice in choices]
        
        # default
        if self.default:
            data['default'] = True
        
        # description
        data['description'] = self.description
        
        # description_localizations
        data['description_localizations'] = destroy_locale_dictionary(self.description_localizations)
        
        type_ = self.type
        
        # max_length
        if (type_ is ApplicationCommandOptionType.string):
            max_length = self.max_length
            if (max_length == 0):
                max_length = APPLICATION_COMMAND_OPTION_MAX_LENGTH_MAX
            
            data['max_length'] = max_length
        
        # max_value
        max_value = self.max_value
        if (max_value is not None):
            data['max_value'] = max_value
        
        # min_length
        if (type_ is ApplicationCommandOptionType.string):
            data['min_length'] = self.min_length
        
        # min_value
        min_value = self.min_value
        if (min_value is not None):
            data['min_value'] = min_value
        
        # name
        data['name'] = self.name
        
        # name_localizations
        data['name_localizations'] = destroy_locale_dictionary(self.name_localizations)
        
        # options
        options = self.options
        if (options is not None):
            data['options'] = [option.to_data() for option in options]
        
        # required
        if self.required:
            data['required'] = True
        
        # type
        data['type'] = type_.value
        
        return data
    
    
    def __repr__(self):
        """Returns the application command option's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # Descriptive fields `.name`, `.description`, `.type`
        
        # name
        repr_parts.append(' name=')
        repr_parts.append(repr(self.name))
        
        # description
        repr_parts.append(', description=')
        repr_parts.append(repr(self.description))
        
        # type
        repr_parts.append(', type=')
        type_ = self.type
        repr_parts.append(repr(type_.value))
        repr_parts.append(' (')
        repr_parts.append(type_.name)
        repr_parts.append(')')
        
        # Extra fields: `.autocomplete`, `.min_value`, `.max_value`, `.default`, `.required`, `.choices`, `.options`
        #    `.name_localizations`, `.description_localizations`, `.min_length`, `.max_length`
        
        # autocomplete
        if self.autocomplete:
            repr_parts.append(', autocomplete=True')
        
        # default
        if self.default:
            repr_parts.append(', default = True')
        
        # required
        if self.required:
            repr_parts.append(', required=True')
        
        if type_ is ApplicationCommandOptionType.string:
            # min_length
            min_length = self.min_length
            if (min_length != 0):
                repr_parts.append(', min_length = ')
                repr_parts.append(repr(min_length))
            
            # max_length
            max_length = self.max_length
            if (max_length != 0):
                repr_parts.append(', max_length = ')
                repr_parts.append(repr(max_length))
        
        
        if type_ is ApplicationCommandOptionType.integer or type_ is ApplicationCommandOptionType.float:
            # min_value
            min_value = self.min_value
            if (min_value is not None):
                repr_parts.append(', min_value = ')
                repr_parts.append(repr(min_value))
            
            # max_value
            max_value = self.max_value
            if (max_value is not None):
                repr_parts.append(', max_value = ')
                repr_parts.append(repr(max_value))
        
        # channel_types
        channel_types = self.channel_types
        if (channel_types is not None):
            repr_parts.append(', channel_types = [')
            
            index = 0
            limit = len(channel_types)
            
            while True:
                channel_type = channel_types[index]
                index += 1
                repr_parts.append(repr(channel_type))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        # choices
        choices = self.choices
        if (choices is not None):
            repr_parts.append(', choices=[')
            
            index = 0
            limit = len(choices)
            
            while True:
                choice = choices[index]
                index += 1
                repr_parts.append(repr(choice))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
        
        # options
        options = self.options
        if (options is not None):
            repr_parts.append(', options=[')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        # name_localizations
        name_localizations = self.name_localizations
        if (name_localizations is not None):
            repr_parts.append(', name_localizations=')
            repr_parts.append(repr(name_localizations))
        
        # description_localizations
        description_localizations = self.description_localizations
        if (description_localizations is not None):
            repr_parts.append(', description_localizations=')
            repr_parts.append(repr(description_localizations))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def copy(self):
        """
        Copies the ``ApplicationCommandOption``.
        
        Returns
        -------
        new : ``ApplicationCommandOption``
        """
        new = object.__new__(type(self))
        
        # autocomplete
        new.autocomplete = self.autocomplete
        
        # channel_types
        channel_types = self.channel_types
        if (channel_types is not None):
            channel_types = tuple(channel_types)
        new.channel_types = channel_types
        
        # choices
        choices = self.choices
        if (choices is not None):
            choices = [choice.copy() for choice in choices]
        new.choices = choices
        
        # default
        new.default = self.default
        
        # description
        new.description = self.description
        
        # description_localizations
        description_localizations = self.description_localizations
        if (description_localizations is not None):
            description_localizations = description_localizations.copy()
        new.description_localizations = description_localizations
        
        # max_length
        new.max_length = self.max_length
        
        # max_value
        new.max_value = self.max_value
        
        # min_length
        new.min_length = self.min_length
        
        # min_value
        new.min_value = self.min_value
        
        # name
        new.name = self.name
        
        # name_localizations
        name_localizations = self.name_localizations
        if (name_localizations is not None):
            name_localizations = name_localizations.copy()
        new.name_localizations = name_localizations
        
        # options
        options = self.options
        if (options is not None):
            options = [option.copy() for option in options]
        new.options = options
        
        # required
        new.required = self.required
        
        # type
        new.type = self.type
        
        return new
    
    
    def __eq__(self, other):
        """Returns whether the two options are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # autocomplete
        if self.autocomplete != other.autocomplete:
            return False
        
        # channel_types
        if self.channel_types != other.channel_types:
            return False
        
        # choices
        if self.choices != other.choices:
            return False
        
        # default
        if self.default != other.default:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # description_localizations
        if self.description_localizations != other.description_localizations:
            return False
        
        # max_length
        if self.max_length != other.max_length:
            return False
        
        # max_value
        if self.max_value != other.max_value:
            return False
        
        # min_length
        if self.min_value != other.min_value:
            return False
        
        # min_value
        if self.min_value != other.min_value:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # name_localizations
        if self.name_localizations != other.name_localizations:
            return False
        
        # options
        if self.options != other.options:
            return False
        
        # required
        if self.required != other.required:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        return True
    
    
    def __len__(self):
        """Returns the application command option's length."""
        length = 0
        
        # choices
        choices = self.choices
        if (choices is not None):
            for choice in choices:
                length += len(choice)
        
        # description & description_localizations
        length += get_localized_length(self.description, self.description_localizations)
        
        # name & name_localizations
        length += get_localized_length(self.name, self.name_localizations)
        
        # options
        options = self.options
        if (options is not None):
            for option in options:
                length += len(option)
        
        return length
    
    
    def __hash__(self):
        """Returns the hash value of the application command option."""
        hash_value = 0
        
        # autocomplete
        hash_value ^= self.autocomplete
        
        # channel_types
        channel_types = self.channel_types
        if (channel_types is not None):
            hash_value ^= hash(channel_types)
        
        # choices
        choices = self.choices
        if (choices is not None):
            hash_value ^= len(choices) << 1
            
            for choice in choices:
                hash_value ^= hash(choice)
        
        # default
        hash_value ^= self.default << 5
        
        # description
        # Do not hash `.description` if equals to `.name`.
        description = self.description
        if (description != self.name):
            hash_value ^= hash(description)
        
        # description_localizations
        # Do not hash `.description_localizations` if equals to `.name_localizations`.
        description_localizations = self.description_localizations
        if (description_localizations is not None) and (description_localizations != self.name_localizations):
            hash_value ^= hash_locale_dictionary(description_localizations)
        
        # max_length
        hash_value ^= self.max_length << 9
        
        # max_value
        hash_value ^= hash(self.max_value)
        
        # min_length
        hash_value ^= self.min_length << 13
        
        # min_value
        hash_value ^= hash(self.min_value)
        
        # name
        hash_value ^= hash(self.name)
        
        # name_localizations
        name_localizations = self.name_localizations
        if (name_localizations is not None):
            hash_value ^= hash_locale_dictionary(name_localizations)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= len(options) << 1
            
            for option in options:
                hash_value ^= hash(option)
        
        # required
        hash_value ^= self.required << 17
        
        # type
        hash_value ^= self.type.value << 18
        
        return hash_value
    
    
    def apply_translation(self, translation_table, replace=False):
        """
        Applies translation from the given nested dictionary to the application command option.
        
        Parameters
        ----------
        translation_table : `None`, `dict` of ((``Locale``, `str`),
                (`None`, `dict` (`str`, (`None`, `str`)) items)) items
            Translation table to pull localization. from.
        replace : `bool` = `False`, Optional
            Whether actual translation should be replaced.
        """
        if translation_table is None:
            return
        
        # choices
        choices = self.choices
        if (choices is not None):
            for choice in choices:
                choice.apply_translation(translation_table, replace)
        
        # description
        self.description_localizations = apply_translation_into(
            self.description,
            self.description_localizations,
            translation_table,
            replace,
        )
        
        # name
        self.name_localizations = apply_translation_into(
            self.name,
            self.name_localizations,
            translation_table,
            replace,
        )
        
        # options
        options = self.options
        if (options is not None):
            for option in options:
                option.apply_translation(translation_table, replace)



__all__ = ('InteractionMetadataFormSubmit',)

import warnings

from scarletio import copy_docs

from .base import InteractionMetadataBase
from .fields import (
    parse_custom_id, parse_components, put_custom_id_into, put_components_into, validate_custom_id, validate_components
)


class InteractionMetadataFormSubmit(InteractionMetadataBase):
    """
    Interaction metadata used when the interaction was triggered by an application command's auto completion.
    
    Parameters
    ----------
    components : `None`, `tuple` of ``InteractionComponent``
        Submitted component values of a form submit interaction.
    
    custom_id : `None`, `str`
        Component or form interaction's custom identifier.
    """
    __slots__ = ('components', 'custom_id')
    
    @classmethod
    @copy_docs(InteractionMetadataBase._create_empty)
    def _create_empty(cls):
        self = object.__new__(cls)
        self.components = None
        self.custom_id = None
        return self
    
    
    @copy_docs(InteractionMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        components = self.components
        if (components is not None):
            components = tuple(option.copy() for option in components)
        new.components = components
        
        new.custom_id = self.custom_id
        
        return new
    
    
    @copy_docs(InteractionMetadataBase._set_attributes_from_keyword_parameters)
    def _set_attributes_from_keyword_parameters(self, keyword_parameters):
        # components
        try:
            components = keyword_parameters.pop('components')
        except KeyError:
            pass
        else:
            self.components = validate_components(components)
        
        # custom_id
        try:
            custom_id = keyword_parameters.pop('custom_id')
        except KeyError:
            pass
        else:
            self.custom_id = validate_custom_id(custom_id)
        
        InteractionMetadataBase._set_attributes_from_keyword_parameters(self, keyword_parameters)
    
    
    @classmethod
    @copy_docs(InteractionMetadataBase.from_data)
    def from_data(cls, data, interaction_event):
        self = object.__new__(cls)
        self.components = parse_components(data)
        self.custom_id = parse_custom_id(data)
        return self
    
    
    @copy_docs(InteractionMetadataBase.to_data)
    def to_data(self, *, defaults = False, interaction_event = None):
        data = {}
        put_components_into(self.components, data, defaults)
        put_custom_id_into(self.custom_id, data, defaults)
        return data
    
    
    @copy_docs(InteractionMetadataBase._put_attribute_representations_into)
    def _put_attribute_representations_into(self, repr_parts):
        
        # components
        components = self.components
        if (components is not None):
            repr_parts.append(' components = ')
            repr_parts.append(repr(components))
            
            field_added = False
        else:
            field_added = True
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' custom_id = ')
            repr_parts.append(repr(custom_id))
        
        return field_added
    
    
    @copy_docs(InteractionMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # components
        components = self.components
        if (components is not None):
            hash_value ^= len(components)
            
            for option in components:
                hash_value ^= hash(option)
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(custom_id)
        
        return hash_value
    
    
    @copy_docs(InteractionMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # components
        if self.components != other.components:
            return False
        
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        return True
    
    
    @property
    def options(self):
        """
        At the case of ``InteractionMetadataFormSubmit`` instances ``.components`` is deprecated and will be removed
        in 2023 February. Please use ``.components``.
        
        If you are accessing this field as a proxied attribute, you can ignore this warning.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.components` is deprecated and will be removed in 2023 February. '
                f'Please use `.components` instead. If you are accessing this field as a proxied attribute, '
                f'you can ignore this warning.'
            )
        )
        
        return self.components



__all__ = ('Oauth2Access', )

from datetime import datetime, timedelta
from time import time as time_now

from scarletio import RichAttributeErrorBaseType

from .helpers import parse_joined_oauth2_scopes
from .preinstanced import Oauth2Scope


class Oauth2Access(RichAttributeErrorBaseType):
    """
    Represents a Discord oauth2 access object, what is returned by ``Client.activate_authorization_code`` if
    activating the authorization code went successfully.
    
    Attributes
    ----------
    access_token : `str`
        Token used for `Bearer` authorizations, when requesting OAuth2 data about the respective user.
    
    created_at : `datetime`
        The time when the access was last created or renewed.
    
    expires_after : `int`
        The time in seconds after this access expires.
    
    redirect_url : `str`
        The redirect url with what the user granted the authorization code for the oauth2 scopes for the application.
        
        Can be empty string if application's owner's access was requested.
    
    refresh_token : `str`
        The token used to renew the access token.
        
        Can be empty string if application's owner's access was requested.
    
    scopes : `None`, `tuple` of ``Oauth2Scope``
        A sequence of the scopes, which the user granted with the access token.
        
        Defaults to `None` if empty.
    """
    __slots__ = ('access_token', 'created_at', 'expires_after', 'redirect_url', 'refresh_token', 'scopes',)
    
    def __init__(self, data, redirect_url):
        """
        Creates an ``Oauth2Access``.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received access data.
        redirect_url : `str`
            The redirect url with what the user granted the authorization code for the oauth2 scopes for the
            application.
        """
        self.redirect_url = redirect_url
        self.access_token = data['access_token']
        self.refresh_token = data.get('refresh_token', '')
        self.expires_after = data['expires_in'] # default is 604800 (s) (1 week)
        self.scopes = parse_joined_oauth2_scopes(data['scope'])
        
        self.created_at = datetime.utcnow() # important for renewing
    
    
    def _renew(self, data):
        """
        Renews the access with the given data.
        
        Parameters
        ----------
        data : `None` or (`dict` of (`str`, `Any`))
            Requested access data.
        """
        self.created_at = datetime.utcnow()
        if data is None:
            return
        
        self.access_token = data['access_token']
        self.refresh_token = data.get('refresh_token', '')
        self.expires_after = data['expires_in']
        self.scopes = parse_joined_oauth2_scopes(data['scope'])
    
    
    @property
    def expires_at(self):
        """
        Returns when the access expires.
        
        Returns
        -------
        expires_at : `datetime`
        """
        return self.created_at + timedelta(seconds=self.expires_after)
    
    
    def __repr__(self):
        """Returns the representation of the achievement."""
        repr_parts = ['<', self.__class__.__name__]
        
        if self.created_at.timestamp() + self.expires_after > time_now():
            state = 'active'
        else:
            state = 'expired'
        
        repr_parts.append(' ')
        repr_parts.append(state)
        
        repr_parts.append('scope count=')
        repr_parts.append(str(len(self.scopes)))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def has_scope(self, scope):
        """
        Returns whether the access has the given scope.
        
        Parameters
        ----------
        scope : ``Oauth2Scope``, `str`
            The scope to check out.
        
        Returns
        -------
        has_scope : `bool`
        
        Raises
        ------
        TypeError
            - If `scope` is neither `str`, ``Oauth2Scope``.
        """
        if isinstance(scope, Oauth2Scope):
            pass
        
        elif isinstance(scope, str):
            scope = Oauth2Scope.get(scope)
        
        else:
            raise TypeError(
                f'`scope` can be `str`, `{Oauth2Scope.__name__}`, got {scope.__class__.__name__}; {scope!r}.'
            )
        
        scopes = self.scopes
        if scopes is None:
            return False
        
        return (scope in scopes)



__all__ = ('ButtonStyle', 'TextInputStyle')

from ...bases import Preinstance as P, PreinstancedBase


class ButtonStyle(PreinstancedBase):
    """
    Represents a button component's style.
    
    Attributes
    ----------
    name : `str`
        The name of the button style.
    value : `int`
        The identifier value the button style
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ButtonStyle``) items
        Stores the predefined ``ButtonStyle``-s. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The button style's type.
    DEFAULT_NAME : `str` = `'UNDEFINED'`
        The default name of the button styles.
    
    Every predefined button style can be accessed as class attribute as well:
    
    +-----------------------+---------------+-------+
    | Class attribute name  | Name          | Value |
    +=======================+===============+=======+
    | none                  | none          | 0     |
    +-----------------------+---------------+-------+
    | blue                  | blue          | 1     |
    +-----------------------+---------------+-------+
    | gray                  | gray          | 2     |
    +-----------------------+---------------+-------+
    | green                 | green         | 3     |
    +-----------------------+---------------+-------+
    | red                   | red           | 4     |
    +-----------------------+---------------+-------+
    | link                  | link          | 5     |
    +-----------------------+---------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    DEFAULT_NAME = 'UNDEFINED'
    
    __slots__ = ()
    
    none = P(0, 'none')
    blue = P(1, 'blue')
    gray = P(2, 'gray')
    green = P(3, 'green')
    red = P(4, 'red')
    link = P(5, 'link')


class TextInputStyle(PreinstancedBase):
    """
    Represents a text input component's type.
    
    Attributes
    ----------
    name : `str`
        The name of the text input style.
    value : `int`
        The identifier value the text input style
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ButtonStyle``) items
        Stores the predefined ``TextInputStyle``-s. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The text input style's type.
    DEFAULT_NAME : `str` = `'UNDEFINED'`
        The default name of the text input styles.
    
    Every predefined text input style can be accessed as class attribute as well:
    
    +-----------------------+---------------+-------+
    | Class attribute name  | Name          | Value |
    +=======================+===============+=======+
    | none                  | none          | 0     |
    +-----------------------+---------------+-------+
    | short                 | short         | 1     |
    +-----------------------+---------------+-------+
    | paragraph             | paragraph     | 2     |
    +-----------------------+---------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    DEFAULT_NAME = 'UNDEFINED'
    
    __slots__ = ()
    
    none = P(0, 'none')
    short = P(1, 'short')
    paragraph = P(2, 'paragraph')



__all__ = (
    'ApplicationCommandParameterConfigurerWrapper', 'ApplicationCommandPermissionOverwriteWrapper', 'CommandWrapper'
)

import reprlib

from scarletio import RichAttributeErrorBaseType, copy_docs, include

from ...discord.application_command import ApplicationCommandPermissionOverwrite, ApplicationCommandOptionType
from ...discord.application_command.application_command_option import (
    _validate_max_length, _validate_min_length
)
from ...discord.client import Client
from ...discord.guild import Guild
from ...discord.preconverters import preconvert_snowflake

from .converters import (
    ANNOTATION_TYPE_TO_OPTION_TYPE, ANNOTATION_TYPE_TO_STR_ANNOTATION, parse_annotation_description,
    parse_annotation_name,  parse_annotation_type_and_choice, postprocess_channel_types, preprocess_channel_types,
    process_max_and_min_value
)


Slasher = include('Slasher')


class CommandWrapper(RichAttributeErrorBaseType):
    """
    Wraps a command enabling the wrapper to postprocess the created command.
    
    Attributes
    ----------
    _wrapped : `Any`
        The wrapped object.
    """
    __slots__ = ('_wrapped',)
    
    def __new__(cls):
        """
        Creates a partial function to wrap a command.
        
        Subclasses should overwrite this method.
        """
        self = RichAttributeErrorBaseType.__new__(cls)
        
        self._wrapped = None
        
        return self
    
    
    def __call__(self, wrapped):
        """
        Wraps the given command.
        
        Subclasses should overwrite this method.
        
        Parameters
        ----------
        wrapped : `Any`
            The command or other wrapper to wrap.
        
        Raises
        -------
        RuntimeError
            - If `self` is already wrapped.
        
        Returns
        -------
        self : ``CommandWrapper``
        """
        if (self._wrapped is not None):
            raise RuntimeError(
                f'`{self!r}` is already wrapped; got {wrapped!r}.'
            )
        
        self._wrapped = wrapped
        return self
    
    
    def apply(self, command):
        """
        Applies the wrapper's changes on the respective command.
        
        Subclasses should overwrite this method.
        
        Parameters
        ----------
        command : ``CommandBaseApplicationCommand``
        """
        pass
    
    
    def __repr__(self):
        """Returns the command wrapper's representation."""
        return f'<{self.__class__.__name__} wrapped={self._wrapped!r}>'
    
    
    def fetch_function_and_wrappers_back(self):
        """
        Fetches back the source function and all the wrappers, the returns them.
        
        Returns
        -------
        function : `Any`
            The wrapped function.
        wrappers : `list` of ``CommandWrapper``
            The fetched back wrappers.
        """
        wrappers = [self]
        maybe_wrapper = self._wrapped
        while True:
            if isinstance(maybe_wrapper, CommandWrapper):
                wrappers.append(maybe_wrapper)
                maybe_wrapper = maybe_wrapper._wrapped
            else:
                function = maybe_wrapper
                break
        
        wrappers.reverse()
        return function, wrappers


class ApplicationCommandPermissionOverwriteWrapper(CommandWrapper):
    """
    Wraps a command allowing / disallowing it only for the given user or role inside of a guild.
    
    Attributes
    ----------
    _wrapped : `Any`
        The wrapped object.
    _guild_id : `int`
        The guild id where the overwrites should be applied to.
    _permission_overwrite : ``ApplicationCommandPermissionOverwrite``
        The permission overwrite to apply.
    """
    __slots__ = ('_guild_id', '_permission_overwrite')
    
    def __new__(cls, guild, target, allow):
        """
        Creates a partial function to wrap a command.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild's identifier where the overwrite is applied.
        target : ``ClientUserBase``, ``Role``, ``Channel``, `tuple` ((``ClientUserBase``, ``Role``, \
                ``Channel``, `str` (`'Role'`, `'role'`, `'User'`, `'user'`, `'Channel'`, `'channel'`)), `int`)
            The target entity of the overwrite
            
            The expected type & value might be pretty confusing, but the target was it to allow relaxing creation.
            To avoid confusing, here is a list of the expected structures:
            
            - ``Role``
            - ``ClientUserBase``
            - ``Channel``
            - `tuple` (``Role``, `int`)
            - `tuple` (``ClientUserBase``, `int`)
            - `tuple` (``Channel``, `int`)
            - `tuple` (`'Role'`, `int`)
            - `tuple` (`'role'`, `int`)
            - `tuple` (`'User'`, `int`)
            - `tuple` (`'user'`, `int`)
            - `tuple` (`'Channel'`, `int`)
            - `tuple` (`'channel'`, `int`)
        
        allow : `bool`
            Whether the respective application command should be enabled for the respective entity.
        """
        if isinstance(guild, Guild):
            guild_id = guild.id
        elif isinstance(guild, (int, str)):
            guild_id = preconvert_snowflake(guild, 'guild')
        else:
            raise TypeError(
                f'`guild` can be `{Guild.__class__.__name__}`, `int`, got {guild.__class__.__name__}; {guild!r}.'
            )
        
        permission_overwrite = ApplicationCommandPermissionOverwrite(target, allow)
        
        self = CommandWrapper.__new__(cls)
        
        self._guild_id = guild_id
        self._permission_overwrite = permission_overwrite
        
        return self
    
    
    def apply(self, command):
        """
        Applies the wrapper's changes on the respective command.
        
        Parameters
        ----------
        command : ``CommandBaseApplicationCommand``
        """
        command.add_permission_overwrite(self._guild_id, self._permission_overwrite)
    
    
    def __repr__(self):
        """Returns the command wrapper's representation."""
        return (
            f'<{self.__class__.__name__} wrapped={self._wrapped!r}, guild_id={self._guild_id!r}, '
            f'overwrite={self._permission_overwrite!r}>'
        )
    
    
    def __matmul__(self, other):
        """Calls the wrapper to wrap a client applying self globally. Can also be used to just apply it to `None`"""
        return self._apply_to(other)
    
    
    @copy_docs(__matmul__)
    def __rmatmul__(self, other):
        return self._apply_to(other)
    
    
    def _apply_to(self, other):
        """
        Tries to apply self to to other.
        
        Parameters
        ----------
        other : `None`, ``ApplicationCommandPermissionOverwriteWrapper``, ``Client``, ``Slasher``
            The object to apply self to.
        
        Returns
        -------
        other : `other`, `self`, `NotImplemented`
            Returns `NotImplemented` if `other`'s type is incompatible.
        
        Raises
        ------
        RuntimeError
            - If `other` is a ``Client`` and it has no `slash` extension setupped.
        """
        if (other is None):
            return self
        
        if type(self) is type(other):
            return self(other)
        
        if isinstance(other, (Client, Slasher)):
            return self._apply_globally(other)
        
        return NotImplemented
    
    
    def _apply_globally(self, other):
        """
        Applies the permission globally to all commands to a client.
        
        Here is an example which enables all commands in the guild only for the given role:
        ```py
        client@set_permission(guild_id, ('role', guild_id), False)
        client@set_permission(guild_id, ('role', role_id), True)
        ```
        
        Parameters
        ----------
        other : ``Client``, ``Slasher``
            The object to add the permission overwrite to.
        
        Returns
        -------
        other : `other`
        
        Raises
        ------
        RuntimeError
            - If `other` has no `slash` extension setupped on it.
        """
        if isinstance(other, Client):
            slasher = getattr(other, 'slasher', None)
            if other is None:
                raise RuntimeError(
                    f'Client {other!r} has no slash extension setupped.'
                )
            
        else:
            slasher = other
        
        slasher._add_permission_overwrites_for_guild(self._guild_id, self._permission_overwrite)
        return other


class ApplicationCommandParameterConfigurerWrapper(CommandWrapper):
    """
    Wraps a command enabling you to modify it's parameter's annotations.
    
    Attributes
    ----------
    _wrapped : `Any`
        The command or other wrapper to wrap.
    _autocomplete : `None`, `CoroutineFunction`
        Auto complete function for the parameter.
    _channel_types : `None`, `tuple` of `int`
        The accepted channel types.
    _choice_enum_type : `None`, `type`
        Enum type of `choices` if applicable.
    _choices : `None`, `dict` of (`str`, `int`, `str`) items
        Parameter's choices.
    _description : `None`, `str`
        Parameter's description.
    _max_length : `int`
        The maximum input length allowed for this option.
    _max_value : `None`, `int`, `float`
        The maximal accepted value by the parameter.
    _min_length : `int`
        The minimum input length allowed for this option.
    _min_value : `None`, `int`, `float`
        The minimal accepted value by the parameter.
    _name : `str`
        The parameter's name.
    _parameter_name : `str`
        The parameter's internal name.
    _type : `int`
        The parameter's internal type identifier.
    """
    __slots__ = (
        '_autocomplete', '_channel_types', '_choice_enum_type', '_choices', '_description', '_max_length',
        '_max_value', '_min_length', '_min_value', '_name', '_parameter_name', '_type'
    )
    
    def __new__(cls, parameter_name, type_or_choice, description = None, name=None, *, autocomplete=None,
            channel_types = None, max_length = None, max_value = None, min_length = None, min_value = None):
        """
        Creates a partial function to wrap a command.
        
        Parameters
        ----------
        parameter_name : `str`
            The parameter's name to modify.
        type_or_choice : `str`, `type`, `list`, `dict`
            The annotation's value to use.
        description : `None`, `str` = `None`, Optional
            Description for the annotation.
        name : `None`, `str` = `None`, Optional
            Name to use instead of the parameter's.
        autocomplete : `None`, `CoroutineFunction` = `None`, Optional (Keyword only)
            Auto complete function for the parameter.
        channel_types : `None`, `iterable` of `int` = `None`, Optional (Keyword only)
            The accepted channel types.
        max_length : `None`, `int` = `None`, Optional (Keyword only)
            The maximum input length allowed for this option.
        max_value : `None`, `int`, `float` = `None`, Optional (Keyword only)
            The maximal accepted value by the parameter.
        min_length : `None`, `int` = `None`, Optional (Keyword only)
            The minimum input length allowed for this option.
        min_value : `None`, `int`, `float` = `None`, Optional (Keyword only)
            The minimal accepted value by the parameter.
        
        Returns
        -------
        wrapper : `functools.partial` of ``CommandWrapper._decorate``
            Partial function to wrap a command.
        
        Raises
        ------
        TypeError
            - If `description`'s is not `None` nor `str`.
            - If `parameter_type_or_choice` is `list`, but it's elements do not match the `tuple`
                (`str`, `str`, `int`) pattern.
            - If `parameter_type_or_choice` is `dict`, but it's items do not match the (`str`, `str`, `int`)
                pattern.
            - If `parameter_type_or_choice` is unexpected.
            - If `name`'s is neither `None`, `str`.
            - If `channel_types` is neither `None` nor `iterable` of `int`.
        ValueError
            - If `description`'s length is out of the expected range [2:100].
            - If `parameter_type_or_choice` is `str`, but not any of the expected ones.
            - If `parameter_type_or_choice` is `type`, but not any of the expected ones.
            - If `choice` amount is out of the expected range [1:25].
            - If `type_or_choice` is a choice, and a `choice` name is duped.
            - If `type_or_choice` is a choice, and a `choice` values are mixed types.
            - If received `channel_types` from both `type_or_choice` and `channel_types` parameters.
        """
        if type(parameter_name) is str:
            pass
        elif isinstance(parameter_name, str):
            parameter_name = str(parameter_name)
        else:
            raise TypeError(
                f'`parameter_name` can be `str`, got {parameter_name.__class__.__name__}; {parameter_name!r}.'
            )
        
        type_, choice_enum_type, choices, parsed_channel_types = parse_annotation_type_and_choice(
            type_or_choice, parameter_name
        )
        
        max_length = _validate_max_length(max_length, ANNOTATION_TYPE_TO_OPTION_TYPE[type_])
        min_length = _validate_min_length(min_length, ANNOTATION_TYPE_TO_OPTION_TYPE[type_])
        
        type_, max_value = process_max_and_min_value(type_, max_value, 'max_value')
        type_, min_value = process_max_and_min_value(type_, min_value, 'min_value')
        
        processed_channel_types = preprocess_channel_types(channel_types)
        channel_types = postprocess_channel_types(processed_channel_types, parsed_channel_types)
        
        if (description is not None):
            description = parse_annotation_description(description, parameter_name)
        
        name = parse_annotation_name(name, parameter_name)
        
        self = CommandWrapper.__new__(cls)
        
        self._autocomplete = autocomplete
        self._choice_enum_type = choice_enum_type
        self._choices = choices
        self._description = description
        self._name = name
        self._parameter_name = parameter_name
        self._type = type_
        self._channel_types = channel_types
        self._max_value = max_value
        self._min_value = min_value
        self._max_length = max_length
        self._min_length = min_length
        
        return self
    
    
    def __repr__(self):
        """Returns the command wrapper's representation."""
        repr_parts = [
            '<', self.__class__.__name__,
            ' wrapped = ', repr(self._wrapped),
            ', parameter_name = ', repr(self.parameter_name),
        ]
        
        type_ = self._type
        type_name = ANNOTATION_TYPE_TO_STR_ANNOTATION[type_]
        repr_parts.append(', type = ')
        repr_parts.append(type_name)
        repr_parts.append(' (')
        repr_parts.append(repr(type_))
        repr_parts.append(')')
        
        autocomplete = self._autocomplete
        if (autocomplete is not None):
            repr_parts.append(', autocomplete = ')
            repr_parts.append(repr(autocomplete))
        
        choice_enum_type = self._choice_enum_type
        if (choice_enum_type is not None):
            repr_parts.append(', choice_enum_type = ')
            repr_parts.append(choice_enum_type.__name__)
        
        choices = self.choices
        if (choices is not None):
            repr_parts.append(', choices = ')
            repr_parts.append(repr(choices))
        
        description = self._description
        if (description is not None):
            repr_parts.append(', description = ')
            repr_parts.append(reprlib.repr(description))
        
        name = self._name
        if (name is not None):
            repr_parts.append(', name = ')
            repr_parts.append(repr(name))
        
        channel_types = self.channel_types
        if (channel_types is not None):
            repr_parts.append(', channel_types = ')
            repr_parts.append(repr(channel_types))
        
        if type_ is ApplicationCommandOptionType.string:
            # min_length
            min_length = self._min_length
            if (min_length != 0):
                repr_parts.append(', min_length = ')
                repr_parts.append(repr(min_length))
            
            # max_length
            max_length = self._max_length
            if (max_length != 0):
                repr_parts.append(', max_length = ')
                repr_parts.append(repr(max_length))
        
        
        if type_ is ApplicationCommandOptionType.integer or type_ is ApplicationCommandOptionType.float:
            min_value = self._min_value
            if (min_value is not None):
                repr_parts.append(', min_value = ')
                repr_parts.append(repr(min_value))
            
            max_value = self._max_value
            if (max_value is not None):
                repr_parts.append(', max_value = ')
                repr_parts.append(repr(max_value))
        
        repr_parts.append('>')
        return ''.join(repr_parts)


def get_parameter_configurers(wrappers):
    """
    Gets the parameter configure wrappers.
    
    Parameters
    ----------
    wrappers : `None`, `list` of ``CommandWrapper``
        The fetched back wrappers if any.
    
    Returns
    -------
    parameter_configurers : `None`, `dict` of (`str`, ``ApplicationCommandParameterConfigurerWrapper``) items
        The collected parameter configurers if any.
    """
    parameter_configurers = None
    
    if (wrappers is not None):
        for wrapper in wrappers:
            if isinstance(wrapper, ApplicationCommandParameterConfigurerWrapper):
                if parameter_configurers is None:
                    parameter_configurers = {}
                
                parameter_configurers[wrapper._parameter_name] = wrapper
                continue
    
    return parameter_configurers



from .import_overwrite import *
from .snapshot import *
from .utils import *

from .client_extension import *
from .constants import *
from .exceptions import *
from .plugin import *
from .plugin_loader import *
from .plugin_root import *
from .helpers import *

__all__ = (
    *import_overwrite.__all__,
    *snapshot.__all__,
    *utils.__all__,
    
    *client_extension.__all__,
    *constants.__all__,
    *exceptions.__all__,
    *plugin.__all__,
    *plugin_loader.__all__,
    *plugin_root.__all__,
    *helpers.__all__,
)

from .. import register_library_extension
register_library_extension('HuyaneMatsu.plugin_loader')
del register_library_extension

import warnings


def __getattr__(attribute_name):
    new_attribute_name = attribute_name.replace(
        'extension', 'plugin'
    ).replace(
        'EXTENSION', 'PLUGIN'
    ).replace(
        'Extension', 'Plugin'
    )
    
    warnings.warn(
        (
            f'`plugin_loader.{attribute_name}` is deprecated and will be removed in 2022 December, '
            f'please use `.{new_attribute_name}` instead.'
        ),
        FutureWarning,
        stacklevel = 2,
    )
    
    # this import is fine, don not worry.
    from .. import plugin_loader
    return getattr(plugin_loader, new_attribute_name)



__all__ = ('ReactionMappingLine',)

from scarletio import include
from scarletio.utils.compact import NEEDS_DUMMY_INIT

Client = include('Client')
ClientUserBase = include('ClientUserBase')


def _validate_reaction_mapping_line_initialise_with(reaction_mapping_line_type, initialize_with):
    """
    Validates the `initialize_with` parameter of ``ReactionMappingLine``.
    
    Parameters
    ----------
    reaction_mapping_line_type : `type`
        The type of the reaction mapping line.
    initialize_with : `None`, `iterable` of (`None`, ``ClientUserBase``), `instance<reaction_mapping_line_type>`
        The value ot initialize the line with.
    
    Returns
    -------
    users : `None`, `list` of ``ClientUserBase``
        The built users from the `initialize_with` parameter.
    unknown : `int`
        The amount of unknown users.
    
    Raises
    ------
    TypeError
        - If an element of `initialize_with` is not `None`, ``ClientUserBase``.
    """
    if initialize_with is None:
        built_initialize_with = None, 0
    
    elif type(initialize_with) is reaction_mapping_line_type:
        built_initialize_with = _validate_reaction_mapping_line_initialise_with_same(initialize_with)
    
    elif (getattr(initialize_with, '__iter__', None) is not None):
        built_initialize_with = _validate_reaction_mapping_line_initialise_with_iterable(initialize_with)
    
    else:
        raise TypeError(
            f'`initialize_with` can be `None`, `{reaction_mapping_line_type.__name__}`, '
            f'`iterable` of (`{ClientUserBase.__name__}`, `None`), got '
            f'{initialize_with.__class__.__name__}; {initialize_with!r}.'
        )
    
    return built_initialize_with


def _validate_reaction_mapping_line_initialise_with_same(initialize_with):
    """
    Validates the `initialize_with` parameter of ``ReactionMappingLine`` where it is the same type.
    
    Parameters
    ----------
    initialize_with : ``ReactionMappingLine``
        The value ot initialize the line with.
    
    Returns
    -------
    users : `None`, `list` of ``ClientUserBase``
        The built users from the `initialize_with` parameter.
    unknown : `int`
        The amount of unknown users.
    """
    if set.__len__(initialize_with) == 0:
        users = None
    else:
        users = [*set.__iter__(initialize_with)]
    
    unknown = initialize_with.unknown
    
    return users, unknown


def _validate_reaction_mapping_line_initialise_with_iterable(initialize_with):
    """
    Validates the `initialize_with` parameter of ``ReactionMappingLine`` where it is an iterable.
    
    Parameters
    ----------
    initialize_with : `iterable` of (`None`, ``ClientUserBase``)
        The value ot initialize the line with.
    
    Returns
    -------
    users : `None`, `list` of ``ClientUserBase``
        The built users from the `initialize_with` parameter.
    unknown : `int`
        The amount of unknown users.
    
    Raises
    ------
    TypeError
        - If an element of `initialize_with` is not `None`, ``ClientUserBase``.
    """
    users = None
    unknown = 0
    
    for user in initialize_with:
        if user is None:
            unknown += 1
            continue
        
        if isinstance(user, ClientUserBase):
            if users is None:
                users = []
            
            users.append(user)
            continue
        
        raise TypeError(
            f'`initialize_with` can contain `{ClientUserBase.__name__}` elements, got '
            f'{user.__class__.__name__}; {user!r}; initialize_with={initialize_with!r}.'
        )
    
    return users, unknown


def _validate_reaction_mapping_line_unknown(unknown):
    """
    Validates the `unknown` parameter of ``ReactionMappingLine``.
    
    Parameters
    ----------
    unknown : `None, `int`
        The given `unknown` value to initialize the reaction line with.
    
    Returns
    -------
    validated_unknown : `int`
    
    Raises
    ------
    TypeError
        - If `unknown` is not `int`, `None`.
    ValueError
        - If `unknown` is negative.
    """
    if unknown is None:
        validated_unknown = 0
    
    elif isinstance(unknown, int):
        if (unknown < 0):
            raise ValueError(
                f'`unknown` cannot be negative, got {unknown!r}.'
            )
        
        validated_unknown = unknown
    
    else:
        raise TypeError(
            f'`unknown` can be `None`, `int`, got {unknown.__class__.__name__}; {unknown!r}.'
        )
    
    return validated_unknown


class ReactionMappingLine(set):
    """
    A `set` subclass which contains the users who reacted with the given ``Emoji`` on a ``Message``.
    
    Attributes
    ----------
    unknown : `int`
        The amount of not known reactors.
    """
    __slots__ = ('unknown',)
    
    if NEEDS_DUMMY_INIT:
        def __init__(self, *args, **kwargs):
            pass
    else:
        __init__ = object.__init__
    
    
    def __new__(cls, initialize_with=None, unknown=None):
        """
        Creates a new reaction mapping line
        
        Parameters
        ----------
        initialize_with : `None`, `iterable` of (`None`, ``ClientUserBase``), `instance<reaction_mapping_line_type>` \
                = `None`, Optional
            The value ot initialize the line with.
        unknown : `None, `int` = `None`, Optional
            The given `unknown` value to initialize the reaction line with.
        
        Raises
        ------
        TypeError
            - If a parameter's type is not acceptable.
        ValueError
            - If a parameter's value is not acceptable.
        """
        validated_line, validated_unknown = _validate_reaction_mapping_line_initialise_with(cls, initialize_with)
        validated_unknown += _validate_reaction_mapping_line_unknown(unknown)
        
        self = set.__new__(cls)
        self.unknown = validated_unknown
        
        if (validated_line is not None):
            set.update(self, validated_line)
        
        return self
    
    
    @classmethod
    def _create_empty(cls, unknown):
        """
        Creates a new reaction mapping line.
        
        Parameters
        ----------
        unknown : `int`
            The amount of not known reactors.
        
        Returns
        -------
        self : instance<cls>
        """
        self = set.__new__(cls)
        self.unknown = unknown
        return self
    
    
    def __len__(self):
        """Returns the amount of users, who reacted with the given emoji on the respective message."""
        return set.__len__(self) + self.unknown
    
    
    def __repr__(self):
        """Returns the representation of the container."""
        repr_parts = [
            self.__class__.__name__,
            '({',
        ]
        
        # set indexing is not public, so we need to do a check, like this
        if set.__len__(self):
            for user in self:
                repr_parts.append(repr(user))
                repr_parts.append(', ')
            
            repr_parts[-1] = '}'
        else:
            repr_parts.append('}')
        
        unknown = self.unknown
        if unknown:
            repr_parts.append(', unknown=')
            repr_parts.append(repr(unknown))
        
        repr_parts.append(')')
        
        return ''.join(repr_parts)
    
    
    def __bool__(self):
        """Returns whether self has any elements."""
        if self.unknown:
            return True
        
        if set.__len__(self):
            return True
        
        return False
    
    
    def __eq__(self, other):
        """Returns whether self equals to other."""
        if type(self) is type(other):
            return self._is_equal_same_type(other)
            
        if (getattr(other, '__iter__', None) is not None):
            return self._is_equal_iterable(other)
        
        return NotImplemented
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether self equals to other. `other` must be same type as self.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other reaction mapping line to compare self to.
        
        Returns
        -------
        is_equal : `bool`
        """
        if self.unknown != other.unknown:
            return False
        
        if not set.__eq__(self, other):
            return False
        
        return True
    
    
    def _is_equal_iterable(self, other):
        """
        Returns whether self equals to other, where other is an iterable.
        
        Parameters
        ----------
        other : `iterable` of (`None`, ``ClientUserBase``)
            The other object to compare self to.
        
        Returns
        -------
        is_equal : `NotImplemented`, `bool`
        """
        unknown = 0
        users = set()
        
        for user in other:
            if user is None:
                unknown += 1
                continue
            
            if isinstance(user, ClientUserBase):
                users.add(user)
                continue
            
            return NotImplemented
        
        
        if self.unknown != unknown:
            return False
        
        if not set.__eq__(self, users):
            return False
        
        return True
    
    
    @classmethod
    def _create_full(cls, users):
        """
        Creates a new ``ReactionMappingLine`` with the given users with `.unknown` set to `0`.
        
        Parameters
        ----------
        users : `list` of ``ClientUserBase``
            A `list`, which should already contain all the users of the reaction mapping line.

        Returns
        -------
        self : ``ReactionMappingLine``
        """
        self = set.__new__(cls)
        self.unknown = 0
        set.update(self, users)
        return self
    
    
    def update(self, users):
        """
        Updates the reaction mapping line with the given users.
        
        Parameters
        ----------
        users : `list` of ``ClientUserBase``
            A `list` of users, who reacted on the respective `Message` with the respective ``Emoji``.
        """
        length_old = len(self)
        set.update(self, users)
        length_new = len(self)
        
        unknown = self.unknown - (length_new - length_old)
        if (unknown < 0):
            unknown = 0
        self.unknown = unknown
    
    
    def copy(self):
        """
        Copies the reaction mapping line.
        
        Returns
        -------
        new : ``ReactionMappingLine``
        """
        new = set.__new__(type(self))
        set.__init__(new, self)
        new.unknown = self.unknown
        return new
    
    
    def filter_after(self, limit, after):
        """
        If we know all the reactors, then instead of executing a Discord API request we filter the reactors locally
        using this method.
        
        Parameters
        ----------
        limit : `int`
            The maximal limit of the users to return.
        after : `int`
            Gets the users after this specified id.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        """
        return sorted([user for user in set.__iter__(self) if user.id > after])[:limit]
    
    
    def clear(self):
        """
        Clears the reaction mapping line by removing every ``User`` object from it.
        """
        clients = []
        for user in self:
            if isinstance(user, Client):
                clients.append(user)

        self.unknown += (set.__len__(self) - len(clients))
        set.clear(self)
        set.update(self, clients)
    
    
    def remove(self, user):
        """
        Removes the given user from self.
        
        Parameters
        ----------
        user : ``ClientUserBase``
            The user to remove.
        
        Returns
        -------
        success : `bool`
            Whether the user was successfully removed.
        """
        try:
            set.remove(self, user)
        except KeyError:
            unknown = self.unknown
            if unknown > 0:
                self.unknown = unknown - 1
                success = True
            
            else:
                success = False
        
        else:
            success = True
        
        return success



__all__ = ('ActivityParty', )

from scarletio import copy_docs

from .base import ActivityFieldBase


def _assert__activity_secrets__id(id_):
    """
    Asserts the `id_` parameter of ``ActivitySecrets.__new__`` method.
    
    Parameters
    ----------
    id_ : `None`, `str`
        The party's id, which in the player is.
    
    Raises
    ------
    AssertionError
        - If `id_` is not `None`, `str`.
    """
    if (id_ is not None) and (not isinstance(id_, str)):
        raise AssertionError(
            f'`id_` can be `None`, `str`, got {id_.__class__.__name__}; {id_!r}.'
        )
    
    return True


def _assert__activity_secrets__size(size):
    """
    Asserts the `size` parameter of ``ActivitySecrets.__new__`` method.
    
    Parameters
    ----------
    size : `int`
        The party's maximal size in which the player is.
    
    Raises
    ------
    AssertionError
        - If `size` is not `int`.
        - If `size` is negative`.
    """
    if (not isinstance(size, int)):
        raise AssertionError(
            f'`size` can be `int`, got {size.__class__.__name__}; {size!r}.'
        )
    
    if (size < 0):
        raise AssertionError(
            f'`size` cannot be negative, got {size!r}.'
        )
    
    return True


def _assert__activity_secrets__max(max_):
    """
    Asserts the `max_` parameter of ``ActivitySecrets.__new__`` method.
    
    Parameters
    ----------
    max_ : `int`
        The party's actual size in which the player is.
    
    Raises
    ------
    AssertionError
        - If `max_` is not `int`.
        - If `max_` is negative`.
    """
    if (not isinstance(max_, int)):
        raise AssertionError(
            f'`max_` can be `int`, got {max_.__class__.__name__}; {max_!r}.'
        )
    
    if (max_ < 0):
        raise AssertionError(
            f'`max_` cannot be negative, got {max_!r}.'
        )
    
    return True


class ActivityParty(ActivityFieldBase):
    """
    Represents a discord activity party.
    
    Attributes
    ----------
    id : `None`, `str`
        The party's id, which in the player is.
    size : `int`
        The party's maximal size in which the player is. Defaults to `0`.
    max : `int`
        The party's actual size in which the player is. Defaults to `0`.
    """
    __slots__ = ('id', 'size', 'max',)
    
    def __new__(cls, *, id_=None, size=0, max_=0):
        """
        Creates a new activity party instance form the given parameters.
        
        Parameters
        ----------
        id_ : `None` = `None`, `str`, Optional (Keyword only)
            The party's id, which in the player is.
        size : `int` = `0`, Optional (Keyword only)
            The party's maximal size, which in the player is.
        max_ : `int` = `0`, Optional (Keyword only)
            The party's actual size, which in the player is.
        """
        assert _assert__activity_secrets__id(id_)
        assert _assert__activity_secrets__size(size)
        assert _assert__activity_secrets__max(max_)
        
        if (id_ is not None) and (not id_):
            id_ = None
        
        self = object.__new__(cls)
        self.id = id_
        self.size = size
        self.max = max_
        return self
    
    
    @classmethod
    @copy_docs(ActivityFieldBase.from_data)
    def from_data(cls, party_data):
        self = object.__new__(cls)
        self.id = party_data.get('id', None)
        
        try:
            size, max_ = party_data['size']
        except KeyError:
            size = 0
            max_ = 0
        
        self.size = size
        self.max = max_
        return self
    
    
    @copy_docs(ActivityFieldBase.to_data)
    def to_data(self):
        party_data = {}
        
        id_ = self.id
        if (id_ is not None):
            party_data['id'] = id_
        
        size = self.size
        max_ = self.max
        if size or max_:
            party_data['size'] = [size, max_]
        
        return party_data
    
    
    @copy_docs(ActivityFieldBase.__repr__)
    def __repr__(self):
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        id_ = self.id
        if (id_ is not None):
            repr_parts.append(' id_=')
            repr_parts.append(repr(id_))
            field_added = True
        else:
            field_added = False
        
        size = self.size
        max_ = self.max
        if size or max_:
            if field_added:
                repr_parts.append(',')
            repr_parts.append(' size=')
            repr_parts.append(repr(size))
            repr_parts.append(', max=')
            repr_parts.append(repr(max_))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    @copy_docs(ActivityFieldBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.id != other.id:
            return False
        
        if self.size != other.size:
            return False
        
        if self.max != other.max:
            return False
        
        return True
    

    @copy_docs(ActivityFieldBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        id_ = self.id
        if (id_ is not None):
            hash_value ^= hash(id_)
            hash_value ^= (1 << 0)
        
        size = self.size
        if size:
            hash_value ^= hash(size)
            hash_value ^= (1 << 4)
        
        max_ = self.max
        if max_:
            hash_value ^= hash(max_)
            hash_value ^= (1 << 8)
        
        return hash_value
    
    
    @copy_docs(ActivityFieldBase.__bool__)
    def __bool__(self):
        id_ = self.id
        if (id_ is not None):
            return True
        
        size = self.size
        if size:
            return True
        
        max_ = self.max
        if max_:
            return True
        
        return False



from .constants import *
from .fields import *
from .preinstanced import *
from .role import *
from .utils import *


__all__ = (
    *constants.__all__,
    *fields.__all__,
    *preinstanced.__all__,
    *role.__all__,
    *utils.__all__,
)



__all__ = ('ComponentMetadataTextInput',)

import reprlib, warnings

from scarletio import copy_docs

from ..shared_fields import parse_custom_id, put_custom_id_into, validate_custom_id

from .base import ComponentMetadataBase
from .constants import MAX_LENGTH_DEFAULT, MIN_LENGTH_DEFAULT, TEXT_INPUT_STYLE_DEFAULT
from .fields import (
    parse_label, parse_max_length, parse_min_length, parse_placeholder, parse_required, parse_text_input_style,
    parse_value, put_label_into, put_max_length_into, put_min_length_into, put_placeholder_into, put_required_into,
    put_text_input_style_into, put_value_into, validate_label, validate_max_length, validate_min_length,
    validate_placeholder, validate_required, validate_text_input_style, validate_value
)


class ComponentMetadataTextInput(ComponentMetadataBase):
    """
    String select component metadata.
    
    Attributes
    ----------
    custom_id : `None`, `str`
        Custom identifier to detect which component was used by the user.
    
    label : `None`, `str`
        Label of the component.
    
    max_length : `int
        The minimal length of the inputted text.
    
    min_length : `int`
        The minimal amount of options to select.
    
    placeholder : `None`, `str`
        Placeholder text of the select.
    
    required : `bool`
        Whether the field is required to be fulfilled.
    
    text_input_style : ``TextInputStyle``
        The text input's style.
    
    value : `None`, `str`
        The text input's default value.
    """
    __slots__ = (
        'custom_id', 'label', 'max_length', 'min_length', 'placeholder', 'required', 'text_input_style', 'value'
    )
    
    @copy_docs(ComponentMetadataBase.__new__)
    def __new__(cls, keyword_parameters):
        # custom_id
        try:
            custom_id = keyword_parameters.pop('custom_id')
        except KeyError:
            custom_id = None
        else:
            custom_id = validate_custom_id(custom_id)
        
        # label
        try:
            label = keyword_parameters.pop('label')
        except KeyError:
            label = None
        else:
            label = validate_label(label)
        
        # max_length
        try:
            max_length = keyword_parameters.pop('max_length')
        except KeyError:
            max_length = MAX_LENGTH_DEFAULT
        else:
            max_length = validate_max_length(max_length)
        
        # min_length
        try:
            min_length = keyword_parameters.pop('min_length')
        except KeyError:
            min_length = MIN_LENGTH_DEFAULT
        else:
            min_length = validate_min_length(min_length)
        
        # placeholder
        try:
            placeholder = keyword_parameters.pop('placeholder')
        except KeyError:
            placeholder = None
        else:
            placeholder = validate_placeholder(placeholder)
        
        # required
        try:
            required = keyword_parameters.pop('required')
        except KeyError:
            required = None
        else:
            if (required is not None):
                required = validate_required(required)
        
        # text_input_style
        try:
            text_input_style = keyword_parameters.pop('text_input_style')
        except KeyError:
            text_input_style = TEXT_INPUT_STYLE_DEFAULT
        else:
            text_input_style = validate_text_input_style(text_input_style)
        
        # value
        try:
            value = keyword_parameters.pop('value')
        except KeyError:
            value = None
        else:
            value = validate_value(value)
        
        # Auto detect required if not-given / None
        
        if (required is None):
            if min_length > 0:
                required = True
            else:
                required = False
        
        # Extra checks
        
        if text_input_style is text_input_style.none:
            text_input_style = TEXT_INPUT_STYLE_DEFAULT
        
        # Construct
        
        self = object.__new__(cls)
        self.custom_id = custom_id
        self.label = label
        self.max_length = max_length
        self.min_length = min_length
        self.placeholder = placeholder
        self.required = required
        self.text_input_style = text_input_style
        self.value = value
        return self
    
    
    @copy_docs(ComponentMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # text_input_style
        text_input_style = self.text_input_style
        repr_parts.append(' text_input_style = ')
        repr_parts.append(text_input_style.name)
        repr_parts.append(' (')
        repr_parts.append(repr(text_input_style.value))
        repr_parts.append(')')
        
        # System fields : custom_id
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            repr_parts.append(', custom_id = ')
            repr_parts.append(reprlib.repr(custom_id))
        
        # Text fields : label & placeholder & value
        
        # label
        label = self.label
        if (label is not None):
            repr_parts.append(', label = ')
            repr_parts.append(reprlib.repr(label))
        
        # placeholder
        placeholder = self.placeholder
        if (placeholder is not None):
            repr_parts.append(', placeholder = ')
            repr_parts.append(repr(placeholder))
        
        # value
        value = self.value
        if (value is not None):
            repr_parts.append(', value=')
            repr_parts.append(repr(value))
        
        # Optional descriptive fields : max_length & min_length & required
        
        # min_length
        min_length = self.min_length
        if min_length:
            repr_parts.append(', min_length = ')
            repr_parts.append(repr(min_length))
        
        # min_length
        max_length = self.max_length
        if max_length:
            repr_parts.append(', max_length = ')
            repr_parts.append(repr(max_length))
        
        # required (relation with `min_length`)
        required = self.required
        if (min_length > 0) ^ required:
            repr_parts.append(', required = ')
            repr_parts.append(repr(required))
        
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ComponentMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(self.custom_id)
        
        # label
        label = self.label
        if (label is not None):
            hash_value ^= hash(label)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= len(options) << 12
            for option in options:
                hash_value ^= hash(option)
        
        # max_length
        max_length = self.max_length
        if (max_length != 1):
            hash_value ^= (max_length << 18)
        
        # min_length
        min_length = self.min_length
        if (min_length != 1):
            min_length ^= (min_length << 22)
        
        # placeholder
        placeholder = self.placeholder
        if (placeholder is not None) and (placeholder != label):
            hash_value ^= hash(placeholder)
        
        # required
        if self.required:
            hash_value ^= (1 << 28)
        
        # text_input_style
        text_input_style = self.text_input_style
        if (text_input_style is not None):
            hash_value ^= text_input_style.value
        
        # value
        value = self.value
        if (value is not None) and (value != label) and (value != placeholder):
            hash_value ^= hash(value)
        
        return hash_value
    
    
    @copy_docs(ComponentMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        # label
        if self.label != other.label:
            return False
        
        # max_length
        if self.max_length != other.max_length:
            return False
        
        # min_length
        if self.min_length != other.min_length:
            return False
        
        # placeholder
        if self.placeholder != other.placeholder:
            return False
        
        # required
        if self.required != other.required:
            return False
        
        # text_input_style
        if self.text_input_style is not other.text_input_style:
            return False
        
        # value
        if self.value != other.value:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ComponentMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.custom_id = parse_custom_id(data)
        self.label = parse_label(data)
        self.max_length = parse_max_length(data)
        self.min_length = parse_min_length(data)
        self.placeholder = parse_placeholder(data)
        self.required = parse_required(data)
        self.text_input_style = parse_text_input_style(data)
        self.value = parse_value(data)
        return self
    
    
    @copy_docs(ComponentMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data =  {}
        
        put_custom_id_into(self.custom_id, data, defaults)
        put_label_into(self.label, data, defaults)
        put_max_length_into(self.max_length, data, defaults)
        put_min_length_into(self.min_length, data, defaults)
        put_placeholder_into(self.placeholder, data, defaults)
        put_required_into(self.required, data, defaults)
        put_text_input_style_into(self.text_input_style, data, defaults)
        put_value_into(self.value, data, defaults)
        
        return data
    
    
    @copy_docs(ComponentMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        new.custom_id = self.custom_id
        new.label = self.label
        new.max_length = self.max_length
        new.min_length = self.min_length
        new.placeholder = self.placeholder
        new.required = self.required
        new.text_input_style = self.text_input_style
        new.value = self.value
        
        return new
    
    
    @copy_docs(ComponentMetadataBase.copy_with)
    def copy_with(self, keyword_parameters):
        # custom_id
        try:
            custom_id = keyword_parameters.pop('custom_id')
        except KeyError:
            custom_id = self.custom_id
        else:
            custom_id = validate_custom_id(custom_id)
        
        # label
        try:
            label = keyword_parameters.pop('label')
        except KeyError:
            label = self.label
        else:
            label = validate_label(label)
        
        # max_length
        try:
            max_length = keyword_parameters.pop('max_length')
        except KeyError:
            max_length = self.max_length
        else:
            max_length = validate_max_length(max_length)
        
        # min_length
        try:
            min_length = keyword_parameters.pop('min_length')
        except KeyError:
            min_length = self.min_length
        else:
            min_length = validate_min_length(min_length)
        
        # placeholder
        try:
            placeholder = keyword_parameters.pop('placeholder')
        except KeyError:
            placeholder = self.placeholder
        else:
            placeholder = validate_placeholder(placeholder)
        
        # required
        try:
            required = keyword_parameters.pop('required')
        except KeyError:
            required = self.required
        else:
            required = validate_required(required)
        
        # text_input_style
        try:
            text_input_style = keyword_parameters.pop('text_input_style')
        except KeyError:
            text_input_style = self.text_input_style
        else:
            text_input_style = validate_text_input_style(text_input_style)

        # value
        try:
            value = keyword_parameters.pop('value')
        except KeyError:
            value = self.value
        else:
            value = validate_value(value)
        
        # Deprecated: style
        
        try:
            style = keyword_parameters.pop('style')
        except KeyError:
            pass
        else:
            warnings.warn(
                (
                    '`style` parameter of components is deprecated and will be removed in 2023 February. '
                    'Please use `text_input_style` for text input components.'
                ),
                FutureWarning,
                stacklevel = 3,
            )
            
            text_input_style = validate_text_input_style(style)
        
        # Extra checks
        
        if text_input_style is text_input_style.none:
            text_input_style = TEXT_INPUT_STYLE_DEFAULT
        
        # Construct
        
        self = object.__new__(type(self))
        self.custom_id = custom_id
        self.label = label
        self.max_length = max_length
        self.min_length = min_length
        self.placeholder = placeholder
        self.required = required
        self.text_input_style = text_input_style
        self.value = value
        return self



__all__ = ()


from scarletio import Compound

from ...bases import maybe_snowflake
from ...core import DISCOVERY_CATEGORIES
from ...guild import DiscoveryCategory, Guild, GuildDiscovery
from ...http import DiscordHTTPClient, rate_limit_groups

from ..functionality_helpers import DiscoveryCategoryRequestCacher, DiscoveryTermRequestCacher
from ..request_helpers import get_guild_discovery_and_guild_id


class ClientCompoundDiscoveryEndpoints(Compound):
    
    http : DiscordHTTPClient
    
    
    async def guild_discovery_get(self, guild):
        """
        Requests and returns the guild's discovery metadata.
        
        The client must have `manage_guild` permission to execute this method.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``
            The guild what's discovery will be requested.
        
        Returns
        -------
        guild_discovery : ``GuildDiscovery``
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_discovery_data = await self.http.guild_discovery_get(guild.id)
        return GuildDiscovery(guild_discovery_data, guild)
    
    
    async def guild_discovery_edit(self, guild, *, primary_category=..., keywords=..., emoji_discovery=...):
        """
        Edits the guild's discovery metadata.
        
        The client must have `manage_guild` permission to execute this method.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, ``GuildDiscovery``, `int`
            The guild what's discovery metadata will be edited or an existing discovery metadata object.
        primary_category : `None`, ``DiscoveryCategory``, `int`, Optional (Keyword only)
            The guild discovery's new primary category's id. Can be given as a ``DiscoveryCategory`` object as well.
            If given as `None`, then resets the guild discovery's primary category id to it's default, what is `0`.
        keywords : `None`, (`iterable` of `str`), Optional (Keyword only)
            The guild discovery's new keywords. Can be given as `None` to reset to the default value, what is `None`,
            or as an `iterable` of strings.
        emoji_discovery : `None`, `bool`, Optional (Keyword only)
            Whether the guild info should be shown when the respective guild's emojis are clicked. If passed as `None`
            then will reset the guild discovery's `emoji_discovery` value to it's default, what is `True`.
        
        Returns
        -------
        guild_discovery : ``GuildDiscovery``
            Updated guild discovery object.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        TypeError
            - If `guild` was neither given as ``Guild``, ``GuildDiscovery``, `int`.
            - If `primary_category_id` was not given neither as `None`, `int`, ``DiscoveryCategory``.
            - If `keywords` was not passed neither as `None`, `iterable` of `str`.
            - If `emoji_discovery` was not given as `None`, `bool`.
        ValueError
            - If `primary_category_id` was given as not primary ``DiscoveryCategory`` object.
            - If `emoji_discovery` was given as `int`, but not as `0`, `1`.
        DiscordException
            If any exception was received from the Discord API.
        """
        if isinstance(guild, Guild):
            guild_id = guild.id
        
        elif isinstance(guild, GuildDiscovery):
            guild_id = guild.guild.id
        
        else:
            guild_id = maybe_snowflake(guild)
            if guild_id is None:
                raise TypeError(
                    f'`guild` can be `{Guild.__name__}`, `{GuildDiscovery.__name__}`, `int`, got '
                    f'{guild.__class__.__name__}; {guild!r}.'
                )
        
        data = {}
        
        if (primary_category is not ...):
            if (primary_category is None):
                primary_category_id = None
            else:
                primary_category_type = primary_category.__class__
                if primary_category_type is DiscoveryCategory:
                    # If name is set means that we should know whether the category is loaded, or just it's `.id`
                    # is known.
                    if (primary_category.name and (not primary_category.primary)):
                        raise ValueError(
                            f'The given `primary_category_id` is not a primary `{DiscoveryCategory.__name__}`, '
                            f'got {primary_category!r}.'
                        )
                    primary_category_id = primary_category.id
                
                elif primary_category_type is int:
                    primary_category_id = primary_category
                
                elif issubclass(primary_category_type, int):
                    primary_category_id = int(primary_category)
                
                else:
                    raise TypeError(
                        f'`primary_category` can be `None`, `int`, `{DiscoveryCategory.__name__}`, '
                        f'got {primary_category_type.__name__}; {primary_category!r}.'
                    )
            
            data['primary_category_id'] = primary_category_id
        
        if (keywords is not ...):
            if (keywords is None):
                pass
            elif (not isinstance(keywords, str)) and hasattr(type(keywords), '__iter__'):
                keywords_processed = set()
                index = 0
                for keyword in keywords:
                    if (type(keyword) is str):
                        pass
                    
                    elif isinstance(keyword, str):
                        keyword = str(keyword)
                    
                    else:
                        raise TypeError(
                            f'`keywords[{index}]` is not `str`, got {keyword.__class__.__name__}; {keyword!r}; '
                            f'keywords={keyword!r}.'
                        )
                    
                    keywords_processed.add(keyword)
                    index += 1
                
                keywords = keywords_processed
            else:
                raise TypeError(
                    f'`keywords` can be `None`, `iterable` of `str`, got {keywords.__class__.__name__}; {keywords!r}.'
                )
            
            data['keywords'] = keywords
        
        if (emoji_discovery is not ...):
            if (emoji_discovery is not None) and (not isinstance(emoji_discovery, bool)):
                raise TypeError(
                    f'`emoji_discovery` can be `None`, `bool`, got {emoji_discovery.__class__.__name__}; '
                    f'{emoji_discovery!r}.'
                )
            
            data['emoji_discoverability_enabled'] = emoji_discovery
        
        guild_discovery_data = await self.http.guild_discovery_edit(guild_id, data)
        
        if isinstance(guild, Guild):
            guild_discovery = GuildDiscovery(guild_discovery_data, guild)
        
        else:
            guild_discovery = guild
            guild_discovery._update_attributes(guild_discovery_data)
        
        return guild_discovery
    
    
    async def guild_discovery_add_subcategory(self, guild, category):
        """
        Adds a discovery subcategory to the guild.
        
        The client must have `manage_guild` permission to execute this method.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, ``GuildDiscovery``, `int`
            The guild to what the discovery subcategory will be added.
        category : ``DiscoveryCategory``, `int`
            The discovery category or it's id what will be added as a subcategory.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild``, ``GuildDiscovery``, neither as `int`.
            - If `category` was not passed neither as ``DiscoveryCategory``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        A guild can have maximum `5` discovery subcategories.
        
        If `guild` was given as ``GuildDiscovery``, then it will be updated.
        """
        guild_discovery, guild_id = get_guild_discovery_and_guild_id(guild)
        
        category_type = category.__class__
        if category_type is DiscoveryCategory:
            category_id = category.id
        elif category_type is int:
            category_id = category
        elif issubclass(category_type, int):
            category_id = int(category)
        else:
            raise TypeError(
                f'`category` can be `int`, `{DiscoveryCategory.__name__}`, got {category_type.__name__}; {category!r}.'
            )
        
        await self.http.guild_discovery_add_subcategory(guild_id, category_id)
        
        if (guild_discovery is not None):
            guild_discovery.sub_categories.add(category_id)
    
    
    async def guild_discovery_delete_subcategory(self, guild, category):
        """
        Removes a discovery subcategory of the guild.
        
        The client must have `manage_guild` permission to execute this method.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, ``GuildDiscovery``, `int`
            The guild to what the discovery subcategory will be removed from.
        category : ``DiscoveryCategory``, `int`
            The discovery category or it's id what will be removed from the subcategories.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild``, ``GuildDiscovery``, neither as `int`.
            - If `category` was not passed neither as ``DiscoveryCategory``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        A guild can have maximum `5` discovery subcategories.
        
        If `guild` was given as ``GuildDiscovery``, then it will be updated.
        """
        guild_discovery, guild_id = get_guild_discovery_and_guild_id(guild)
        
        category_type = category.__class__
        if category_type is DiscoveryCategory:
            category_id = category.id
        elif category_type is int:
            category_id = category
        elif issubclass(category_type, int):
            category_id = int(category)
        else:
            raise TypeError(
                f'`category` can be `int`, `{DiscoveryCategory.__name__}`, got {category_type.__name__}; {category!r}.'
            )
        
        await self.http.guild_discovery_delete_subcategory(guild_id, category_id)
        
        if (guild_discovery is not None):
            guild_discovery.sub_categories.discard(category_id)
    
    
    async def _discovery_category_get_all(self):
        """
        Returns a list of discovery categories, which can be used when editing guild discovery.
        
        This method is a coroutine.
        
        Returns
        -------
        discovery_categories : `list` of ``DiscoveryCategory``
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        discovery_category_datas = await self.http.discovery_category_get_all()
        return [DiscoveryCategory.from_data(discovery_category_data)
            for discovery_category_data in discovery_category_datas]
    
    
    # Add cached, so even tho the first request fails with `ConnectionError` will not be raised.
    discovery_category_get_all = DiscoveryCategoryRequestCacher(
        _discovery_category_get_all,
        3600.0,
        list(DISCOVERY_CATEGORIES.values()),
    )
    
    
    async def discovery_validate_term(self, term):
        """
        Checks whether the given discovery search term is valid.
        
        This method is a coroutine.
        
        Parameters
        ----------
        term : `str`
        
        Returns
        -------
        valid : `bool`
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        data = await self.http.discovery_validate_term({'term': term})
        return data['valid']
    
    discovery_validate_term = DiscoveryTermRequestCacher(
        discovery_validate_term,
        86400.0,
        rate_limit_groups.discovery_validate_term,
    )



__all__ = ('AutoModerationRuleTriggerMetadataKeyword',)

from scarletio import copy_docs

from .base import AutoModerationRuleTriggerMetadataBase


class AutoModerationRuleTriggerMetadataKeyword(AutoModerationRuleTriggerMetadataBase):
    """
    Keyword trigger metadata for an auto moderation rule.
    
    Attributes
    ----------
    keywords : `None`, `tuple` of `str`
        Substrings which will be searched for in content.
    
    Matching Strategies
    -------------------
    Use the `*` wildcard symbol at the beginning and end of a keyword to define how the keyword should be matched.
    All keywords are not case sensitive.
    
    - `Prefix` - word must start with the keyword
        +-----------+---------------------------------------+
        | Keyword   | Matches                               |
        +===========+=======================================+
        | cat*      | catch, Catapult, CAttLE               |
        +-----------+---------------------------------------+
        | tra*      | train, trade, TRAditional             |
        +-----------+---------------------------------------+
        | the mat*  | the matrix                            |
        +-----------+---------------------------------------+
    
    
    - `Suffix` - word must end with the keyword
    
        +-----------+---------------------------------------+
        | Keyword   | Matches                               |
        +===========+=======================================+
        | *cat      | wildcat, copyCat                      |
        +-----------+---------------------------------------+
        | *tra      | extra, ultra, orchesTRA               |
        +-----------+---------------------------------------+
        | *the mat  | breathe mat                           |
        +-----------+---------------------------------------+
    
    
    - `Anywhere` - keyword can appear anywhere in the content
    
        +-----------+---------------------------------------+
        | Keyword     | Matches                             |
        +===========+=======================================+
        | *cat*       | location, eduCation                 |
        +-----------+---------------------------------------+
        | *tra*       | abstracted, outrage                 |
        +-----------+---------------------------------------+
        | *the mat*   | breathe matter                      |
        +-----------+---------------------------------------+
    
    
    - `Whole Word` - keyword is a full word or phrase and must be surrounded by whitespace at the beginning and end
    
        +-----------+---------------------------------------+
        | Keyword   | Matches                               |
        +===========+=======================================+
        | cat       | cat                                   |
        +-----------+---------------------------------------+
        | train     | train                                 |
        +-----------+---------------------------------------+
        | the mat   | the mat                               |
        +-----------+---------------------------------------+
    """
    __slots__ = ('keywords',)
    
    def __new__(cls, keywords):
        """
        Creates a new keyword trigger metadata for ``AutoModerationRule``-s.
        
        Parameters
        ----------
        keywords : `None`, `str`, `iterable` of `str`
            Substrings which will be searched for in content.
               
        Raises
        ------
        TypeError
            - If `keywords` type is incorrect.
            - If a keyword's type is incorrect.
        """
        if keywords is None:
            processed_keywords = None
        
        elif isinstance(keywords, str):
            processed_keywords = (keywords, )
        
        else:
            iterator = getattr(type(keywords), '__iter__', None)
            if iterator is None:
                raise TypeError(
                    f'`keywords` can be `None`, `str`, `iterable` of `str`, '
                    f'got {keywords.__class__.__name__}; {keywords!r}.'
                )
            
            processed_keywords = None
            
            for keyword in iterator(keywords):
                if not isinstance(keyword, str):
                    raise TypeError(
                        f'`keywords` can contain `str` elements, got {keyword.__class__.__name__}; {keyword!r}; '
                        f'keywords={keywords!r}'
                    )
                
                if processed_keywords is None:
                    processed_keywords = set()
                
                processed_keywords.add(keyword)
            
            if (processed_keywords is not None):
                processed_keywords = tuple(sorted(processed_keywords))
        
        self = object.__new__(cls)
        self.keywords = processed_keywords
        return self
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # keywords
        keywords = self.keywords
        repr_parts.append(' keywords=[')
        if (keywords is not None):
            index = 0
            limit = len(keywords)
            while True:
                keyword = keywords[index]
                repr_parts.append(repr(keyword))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
        
        repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @classmethod
    @copy_docs(AutoModerationRuleTriggerMetadataBase.from_data)
    def from_data(cls, data):
        keyword_array = data.get('keyword_filter', None)
        if (keyword_array is None) or (not keyword_array):
            keywords = None
        else:
            keywords = tuple(sorted(keyword_array))
        
        self = object.__new__(cls)
        self.keywords = keywords
        return self
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.to_data)
    def to_data(self):
        data = {}
        
        data['keyword_filter'] = [*self.iter_keywords()]
        
        return data
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.keywords != other.keywords:
            return False
        
        return True
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        keywords = self.keywords
        if (keywords is not None):
            hash_value ^= len(keywords)
            
            for keyword in keywords:
                hash_value ^= hash(keyword)
        
        return hash_value
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.copy)
    def copy(self):
        new = AutoModerationRuleTriggerMetadataBase.copy(self)
        
        # keywords
        keywords = self.keywords
        if (keywords is not None):
            keywords = tuple(keyword for keyword in keywords)
        new.keywords = keywords
        
        return new
    
    
    def iter_keywords(self):
        """
        Iterates over the keywords of the keyword trigger metadata.
        
        This method is an iterable generator.
        
        Yields
        ------
        keyword : `str`
        """
        keywords = self.keywords
        if (keywords is not None):
            yield from keywords



__all__ = ('AuditLog', )

from scarletio import RichAttributeErrorBaseType, WeakReferer

from ...application_command import ApplicationCommand
from ...auto_moderation import AutoModerationRule
from ...channel import Channel
from ...integration import Integration
from ...scheduled_event import ScheduledEvent
from ...user import User
from ...webhook import Webhook

from .audit_log_entry import AuditLogEntry


class AuditLog(RichAttributeErrorBaseType):
    """
    Whenever an admin action is performed on the API, an audit log entry is added to the respective guild's audit
    logs. This class represents a requested  collections of these entries.
    
    Attributes
    ----------
    _self_reference : `None` or ``WeakReferer`` to ``AuditLog``
        Weak reference to the audit log itself.
    application_commands : `dict` of (`int`, ``ApplicationCommand``) items
        A dictionary that contains the mentioned application commands by the audi log entries. The keys are the `id`-s
        of the application commands, meanwhile the values are the application commands themselves.
    auto_moderation_rules : `dict` of (`int`, ``ApplicationCommand``) items
        A dictionary that contains the auto moderation rules mentioned inside of the audit log entries. The keys
        are the `id`-s of the rules and the values are the rules themselves.
    entries : `list` of ``AuditLogEntry``
        A list of audit log entries that the audit log contains.
    guild : ``Guild``
        The audit logs' respective guild.
    integrations : `dict` of (`int`, ``Integration``) items
        A dictionary that contains the mentioned integrations by the audit log's entries. The keys are the `id`-s of
        the integrations, meanwhile the values are the integrations themselves.
    scheduled_events : `dict` of (`int`, ``ScheduledEvent``) items
        A dictionary containing the scheduled events mentioned inside of the audit logs.
    threads : `dict` of (`int`, ``Channel``) items
        A dictionary containing the mentioned threads inside of the audit logs.
    users : `dict` of (`int`, ``ClientUserBase``) items
        A dictionary that contains the mentioned users by the audit log's entries. The keys are the `id`-s of the
        users, meanwhile the values are the users themselves.
    webhooks : `dict` of (`int`, ``Webhook``) items
        A dictionary that contains the mentioned webhook by the audit log's entries. The keys are the `id`-s of the
        webhooks, meanwhile the values are the values themselves.
    """
    __slots__ = (
        '__weakref__', '_self_reference', 'application_commands', 'auto_moderation_rules', 'entries', 'guild',
        'integrations', 'scheduled_events', 'threads', 'users', 'webhooks'
    )
    
    def __new__(cls, data, guild):
        """
        Creates an ``AuditLog`` from the data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        guild : ``Guild``
            The respective guild of the audit logs.
        """
        self = object.__new__(cls)
        self._self_reference = None
        self.application_commands = {}
        self.auto_moderation_rules = {}
        self.entries = []
        self.guild = guild
        self.integrations = {}
        self.scheduled_events = {}
        self.threads = {}
        self.users = {}
        self.webhooks = {}
        return self
    
    
    def _populate(self, data):
        """
        Populates the audit log entry with the given data.
        
        Parameters
        ----------
        data : `dict` (`str`, `Any`) items
            Audit log data.
        
        Returns
        -------
        populated : `bool`
            Whether any entry was added to the audit log.
        """
        try:
            entry_datas = data['audit_log_entries']
        except KeyError:
            return False
        
        if not entry_datas:
            return False
        
        
        try:
            users_data = data['users']
        except KeyError:
            pass
        else:
            users = self.users
            
            for user_data in users_data:
                user = User.from_data(user_data)
                users[user.id] = user
        
        
        try:
            application_command_datas = data['application_commands']
        except KeyError:
            pass
        else:
            application_commands = self.application_commands
            
            for application_command_data in application_command_datas:
                application_command = ApplicationCommand.from_data(application_command_data)
                application_commands[application_command.id] = application_command
        
        
        try:
            auto_moderation_rule_datas = data['auto_moderation_rules']
        except KeyError:
            pass
        else:
            auto_moderation_rules = self.auto_moderation_rules
            
            for auto_moderation_rule_data in auto_moderation_rule_datas:
                auto_moderation_rule = AutoModerationRule.from_data(auto_moderation_rule_data)
                auto_moderation_rules[auto_moderation_rule.id] = auto_moderation_rule
        
        
        try:
            webhooks_data = data['webhook']
        except KeyError:
            pass
        else:
            webhooks = self.webhooks
            
            for webhook_data in webhooks_data:
                webhook = Webhook(webhook_data)
                webhooks[webhook.id] = webhook
        
        
        try:
            integration_datas = data['integrations']
        except KeyError:
            pass
        else:
            integrations = self.integrations
            
            for integration_data in integration_datas:
                integration = Integration.from_data(integration_data)
                integrations[integration.id] = integration
        
        
        try:
            thread_datas = data['threads']
        except KeyError:
            pass
        else:
            threads = self.threads
            
            for thread_data in thread_datas:
                thread = Channel.from_data(thread_data, None, self.guild.id)
                threads[thread.id] = thread
        
        
        try:
            scheduled_event_datas = data['guild_scheduled_events']
        except KeyError:
            pass
        else:
            scheduled_events = self.scheduled_events
            
            for scheduled_event_data in scheduled_event_datas:
                scheduled_event = ScheduledEvent(scheduled_event_data)
                scheduled_events[scheduled_event.id] = scheduled_event
        
        
        entries = self.entries
        for entry_data in entry_datas:
            entry = AuditLogEntry(entry_data, self)
            if (entry is not None):
                entries.append(entry)
        
        return True
    
    
    def _get_self_reference(self):
        self_reference = self._self_reference
        if (self_reference is None):
            self_reference = WeakReferer(self)
            self._self_reference = self_reference
        
        return self_reference
    
    
    def __iter__(self):
        """Iterates over the audit log's entries."""
        return iter(self.entries)
    
    
    def __reversed__(self):
        """Reversed iterator over the audit log's entries."""
        return reversed(self.entries)
    
    
    def __len__(self):
        """Returns the amount of entries that the audit lgo contain."""
        return len(self.entries)
    
    
    def __getitem__(self,index):
        """Returns the specific audit log entry at the given index."""
        return self.entries.__getitem__(index)
    
    
    def __repr__(self):
        """Returns the representation of the Audit log."""
        return f'<{self.__class__.__name__} of {self.guild.name}, length={len(self.entries)}>'



__all__ = ('AuditLogIterator', )

from ...bases import maybe_snowflake
from ...user import ClientUserBase
from ...utils import now_as_id

from ..guild import Guild
from ..utils import create_partial_guild_from_id

from .audit_log import AuditLog
from .audit_log_entry import AuditLogEntry
from .preinstanced import AuditLogEvent


class AuditLogIterator(AuditLog):
    """
    An async iterator over a guild's audit logs.
    
    Attributes
    ----------
    _self_reference : `None` or ``WeakReferer`` to ``AuditLog``
        Weak reference to the audit log itself.
    application_commands : `dict` of (`int`, ``ApplicationCommand``) items
        A dictionary that contains the mentioned application commands by the audi log entries. The keys are the `id`-s
        of the application commands, meanwhile the values are the application commands themselves.
    auto_moderation_rules : `dict` of (`int`, ``ApplicationCommand``) items
        A dictionary that contains the auto moderation rules mentioned inside of the audit log entries. The keys
        are the `id`-s of the rules and the values are the rules themselves.
    entries : `list` of ``AuditLogEntry``
        A list of audit log entries that the audit log contains.
    guild : ``Guild``
        The audit logs' respective guild.
    integrations : `dict` of (`int`, ``Integration``) items
        A dictionary that contains the mentioned integrations by the audit log's entries. The keys are the `id`-s of
        the integrations, meanwhile the values are the integrations themselves.
    scheduled_events : `dict` of (`int`, ``ScheduledEvent``) items
        A dictionary containing the scheduled events mentioned inside of the audit logs.
    threads : `dict` of (`int`, ``Channel``) items
        A dictionary containing the mentioned threads inside of the audit logs.
    users : `dict` of (`int`, ``ClientUserBase``) items
        A dictionary that contains the mentioned users by the audit log's entries. The keys are the `id`-s of the
        users, meanwhile the values are the users themselves.
    webhooks : `dict` of (`int`, ``Webhook``) items
        A dictionary that contains the mentioned webhook by the audit log's entries. The keys are the `id`-s of the
        webhooks, meanwhile the values are the values themselves.
    
    _data : `dict` of (`str`, `Any`) items
        Data to be sent to Discord when requesting an another audit log chunk. Contains some information, which are not
        stored by any attributes of the audit log iterator, these are the filtering `user` and `event` options.
    _index : `int`
        The next audit log entries index to yield.
    client : ``Client``
        The client, who will execute the api requests.
    """
    __slots__ = ('_data', '_index', 'client',)
    
    async def __new__(cls, client, guild, user=None, event = None):
        """
        Creates an audit log iterator with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client, who will execute the api requests.
        guild : ``Guild``, `int`
            The guild what's audit logs will be requested.
        user : `None`, ``ClientUserBase``, `int` = `None`, Optional
            Whether the audit logs should be filtered only to those, which were created by the given user.
        event : `None`, ``AuditLogEvent``, `int` = `None`, Optional
            Whether the audit logs should be filtered only on the given event.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild``, nor as `int`.
            - If `user` was not given neither as `None`, ``ClientUserBase`` nor as `int`.
            - If `event` as not not given neither as `None`, ``AuditLogEvent`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        data = {
            'limit': 100,
            'before': now_as_id(),
        }
        
        if isinstance(guild, Guild):
            guild_id = guild.id
        else:
            guild_id = maybe_snowflake(guild)
            if guild_id is None:
                raise TypeError(
                    f'`guild_or_discovery` can be `{Guild.__name__}`, `int`, got '
                    f'{guild.__class__.__name__}; {guild!r}.'
                )
            
            guild = None
        
        if (user is not None):
            if isinstance(user, ClientUserBase):
                user_id = user.id
            
            else:
                user_id = maybe_snowflake(user)
                if user_id is None:
                    raise TypeError(
                        f'`user` can be `{ClientUserBase.__name__}`, `int`, got'
                        f'{user.__class__.__name__}; {user!r}.'
                    )
            
            data['user_id'] = user_id
        
        if (event is not None):
            if isinstance(event, AuditLogEvent):
                event_value = event.value
            elif isinstance(event, int):
                event_value = event
            else:
                raise TypeError(
                    f'`event` can be `None`, `{AuditLogEvent.__name__}`, `int`, got '
                    f'{event.__class__.__name__}; {event!r}.'
                )
            
            data['action_type'] = event_value
        
        if guild is None:
            log_data = await client.http.audit_log_get_chunk(guild_id, data)
            if guild is None:
                guild = create_partial_guild_from_id(guild_id)
        else:
            log_data = None
        
        self = AuditLog.__new__(cls, data, guild)
        self._data = data
        self._index = 0
        self.client = client
        
        if (log_data is not None):
            self._populate(log_data)
        
        return self
    
    
    async def load_all(self):
        """
        Loads all not yet loaded audit logs of the audit log iterator's guild.
        
        This method is a coroutine.
        """
        entries = self.entries
        client = self.client
        http = client.http
        data = self._data
        
        while True:
            if entries:
                data['before'] = entries[-1].id
            
            log_data = await http.audit_log_get_chunk(self.guild.id, data)
            
            if not self._populate(log_data):
                return
            
            if len(entries) % 100:
                return
    
    
    def transform(self):
        """
        Converts the audit log iterator to an audit log object.
        
        Returns
        -------
        audit_log : ``AuditLog``
        """
        audit_log = object.__new__(AuditLog)
        audit_log.application_commands = self.application_commands
        audit_log.auto_moderation_rules = self.auto_moderation_rules
        audit_log.entries = self.entries
        audit_log.guild = self.guild
        audit_log.integrations = self.integrations
        audit_log.scheduled_events = self.scheduled_events
        audit_log.threads = self.threads
        audit_log.users = self.users
        audit_log.webhooks = self.webhooks
        return audit_log
    
    
    def __aiter__(self):
        """Returns self and resets the `.index`."""
        self._index = 0
        return self
    
    
    async def __anext__(self):
        """
        Yields the next entry of the audit log iterator.
        
        This method is a coroutine.
        """
        ln = len(self.entries)
        index = self._index
        
        if index < ln:
            self._index += 1
            return self.entries[index]
        
        if index % 100:
            raise StopAsyncIteration
        
        data = self._data
        if ln:
            data['before'] = self.entries[ln - 1].id
        
        log_data = await self.client.http.audit_log_get_chunk(self.guild.id, data)
        
        
        if not self._populate(log_data):
            raise StopAsyncIteration
        
        self._index += 1
        return self.entries[index]



__all__ = ('Achievement', )

from ...env import API_VERSION

from ..bases import DiscordEntity, IconSlot
from ..http import urls as module_urls
from ..localization.utils import build_locale_dictionary
from ..utils import DATETIME_FORMAT_CODE


class Achievement(DiscordEntity):
    """
    Represents a Discord achievement created at Developer portal.
    
    Attributes
    ----------
    id : `int`
        The achievement's unique identifier number.
    application_id : `int`
        The achievement's respective application's id.
    description : `str`
        The description of the achievement.
    description_localizations : `None`, `dict` of (`str`, `str`) items
        Localized descriptions of the achievement.
    icon_hash : `int`
        The achievement's icon's hash. Achievements always have icon.
    icon_type : ``IconType``
        The achievement's icon's type.
    name : `str`
        The name of the achievement.
    name_localizations : `None`, `dict` of (`str`, `str`) items
        Localized names of the achievement.
    secret : `bool`
        Secret achievements will *not* be shown to the user until they've unlocked them.
    secure : `bool`
        Secure achievements can only be set via HTTP calls from your server, not by a game client using the SDK.
    """
    __slots__ = (
        'application_id', 'description', 'description_localizations', 'name', 'name_localizations', 'secret', 'secure'
    )
    
    icon = IconSlot('icon', 'icon_hash', module_urls.achievement_icon_url, module_urls.achievement_icon_url_as)
    
    def __init__(self, data):
        """
        Creates an achievement with the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received achievement data.
        """
        self.application_id = int(data['application_id'])
        self.id = int(data['id'])
        
        self._update_attributes(data)
    
    
    def __repr__(self):
        """Returns the achievement's representation."""
        return f'<{self.__class__.__name__} id={self.id}, name={self.name!r}>'
    
    
    def __format__(self, code):
        """Formats the achievement with the given format code."""
        if not code:
            return self.name
        
        if code == 'c':
            return format(self.created_at, DATETIME_FORMAT_CODE)
        
        raise ValueError(
            f'Unknown format code {code!r} for {self.__class__.__name__}; {self!r}. '
            f'Available format codes: {""!r}, {"c"!r}.'
        )
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the achievement and returns it's overwritten attributes as a `dict` with a `attribute-name` -
        `old-value` relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Achievement data received from Discord.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        
        +---------------------------+-------------------------------------------+
        | Keys                      | Values                                    |
        +===========================+===========================================+
        | description               | `str`                                     |
        +---------------------------+-------------------------------------------+
        | description_localizations | `None`, `dict` of (`str`, `str`) items    |
        +---------------------------+-------------------------------------------+
        | icon                      | ``Icon``                                  |
        +---------------------------+-------------------------------------------+
        | name                      | `str`                                     |
        +---------------------------+-------------------------------------------+
        | name_localizations        | `None`, `dict` of (`str`, `str`) items    |
        +---------------------------+-------------------------------------------+
        | secret                    | `bool`                                    |
        +---------------------------+-------------------------------------------+
        | secure                    | `bool`                                    |
        +---------------------------+-------------------------------------------+
        """
        old_attributes = {}
        
        if API_VERSION >= 10:
            name_localizations = data['name_localizations']
            if not name_localizations:
                name_localizations = None
            
            name = data['name']
            
            description_localizations = data['description_localizations']
            if not description_localizations:
                description_localizations = None
            
            description = data['description']
        
        else:
            name_localizations = data['name']
            name = data.pop('default')
            if not name_localizations:
                name_localizations = None
            
            description_localizations = data['description']
            description = description_localizations.pop('default')
            if not description_localizations:
                description_localizations = None
        
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        
        if self.description != description:
            old_attributes['description'] = self.description
            self.description = description
        
        
        if self.name_localizations != name_localizations:
            old_attributes['name_localizations'] = self.name_localizations
            self.name_localizations = build_locale_dictionary(name_localizations)
        
        
        if self.description_localizations != description_localizations:
            old_attributes['description_localizations'] = self.description_localizations
            self.description_localizations = build_locale_dictionary(description_localizations)
        
        
        secret = data['secret']
        if self.secret != secret:
            old_attributes['secret'] = self.secret
            self.secret = secret
        
        secure = data['secure']
        if self.secure != secure:
            old_attributes['secure'] = self.secure
            self.secure = secure
        
        self._update_icon(data, old_attributes)
        
        return old_attributes
    
    
    def _update_attributes(self, data):
        """
        Updates the achievement with overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Achievement data received from Discord.
        """
        if API_VERSION >= 10:
            name_localizations = data['name_localizations']
            if not name_localizations:
                name_localizations = None
            
            name = data['name']
            
            description_localizations = data['description_localizations']
            if not description_localizations:
                description_localizations = None
            
            description = data['description']
        
        else:
            name_localizations = data['name']
            name = data.pop('default')
            if not name_localizations:
                name_localizations = None
            
            description_localizations = data['description']
            description = description_localizations.pop('default')
            if not description_localizations:
                 description_localizations = None
        
        
        self.name = name
        self.description = description
        
        self.name_localizations = name_localizations
        self.description_localizations = description_localizations
        
        self.secret = data['secret']
        self.secure = data['secure']
        
        self._set_icon(data)



__all__ = ('InteractionOption',)

import reprlib

from scarletio import RichAttributeErrorBaseType, export

from ...application_command import ApplicationCommandOptionType

from .fields import (
    parse_focused, parse_name, parse_options, parse_type, parse_value, put_focused_into, put_name_into,
    put_options_into, put_type_into, put_value_into, validate_focused, validate_name, validate_options, validate_type,
    validate_value
)


@export
class InteractionOption(RichAttributeErrorBaseType):
    """
    Integration option representing a parameter or a sub-command-group.
    
    Attributes
    ----------
    focused : `bool`
        Whether this field is focused by the user. This field is applicable for auto complete options.
    
    name : `str`
        The option's name.
    
    options : `None`, `tuple` of ``InteractionOption``
        Nested options.
    
    type : ``ApplicationCommandOptionType``
        The represented option's type.
    
    value : `None`, `str`, `float`, `int`
        The value the user has been typed or selected.
    """
    __slots__ = ('focused', 'name', 'options', 'type', 'value')
    
    
    def __new__(cls, **keyword_parameters):
        """
        Creates a new interaction option from the given keyword parameters.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            Keyword parameters defining which fields and how should be set.
        
        Other Parameters
        ----------------
        focused : `bool`, Optional (Keyword only)
            Whether this field is focused by the user.
        
        name : `str`, Optional (Keyword only)
            The option's name.
        
        options : `None`, `tuple` of ``InteractionOption``, Optional (Keyword only)
            Nested options.
        
        type_ : ``ApplicationCommandOptionType``, Optional (Keyword only)
            The represented option's type.
        
        value : `None`, `str`, `float`, `int`, Optional (Keyword only)
            The value the user has been typed or selected.
        
        Raises
        ------
        TypeError
            - If a field's type is incorrect.
            - Extra or unused fields given.
        ValueError
            - If a field's value is incorrect.
        """
        self = cls._create_empty()
        self._populate_from_keyword_parameters(keyword_parameters)
        return self
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates an interaction option with it's default values set.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.focused = False
        self.name = ''
        self.options = None
        self.type = ApplicationCommandOptionType.none
        self.value = None
        return self
    
    
    def copy(self):
        """
        Copies the interaction option.
        
        Returns
        -------
        new : `instance<cls>`
        """
        new = object.__new__(type(self))
        new.focused = self.focused
        new.name = self.name
        options = self.options
        if (options is not None):
            options = tuple(option.copy() for option in options)
        new.options = options
        new.type = self.type
        new.value = self.value
        return new
    
    
    def copy_with(self, **keyword_parameters):
        """
        Copies the interaction option with replacing the defined fields.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            Keyword parameters defining which fields and how should be set.
        
        Other Parameters
        ----------------
        focused : `bool`, Optional (Keyword only)
            Whether this field is focused by the user.
        
        name : `str`, Optional (Keyword only)
            The option's name.
        
        options : `None`, `tuple` of ``InteractionOption``, Optional (Keyword only)
            Nested options.
        
        type_ : ``ApplicationCommandOptionType``, Optional (Keyword only)
            The represented option's type.
        
        value : `None`, `str`, `float`, `int`, Optional (Keyword only)
            The value the user has been typed or selected.
        
        Returns
        -------
        new : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a field's type is incorrect.
            - Extra or unused fields given.
        ValueError
            - If a field's value is incorrect.
        """
        self = self.copy()
        self._populate_from_keyword_parameters(keyword_parameters)
        return self
    
    
    def _populate_from_keyword_parameters(self, keyword_parameters):
        """
        Sets the interaction option's attributes from the given keyword parameters.
        
        Parameters
        keyword_parameters : `dict` of (`str`, `object`) items
            A dictionary of keyword parameters defining which fields and how should be set.
        
        Raises
        ------
        TypeError
            - If a field's type is incorrect.
            - Extra or unused fields given.
        ValueError
            - If a field's value is incorrect.
        """
        if not keyword_parameters:
            return
        
        # focused
        try:
            focused = keyword_parameters.pop('focused')
        except KeyError:
            pass
        else:
            self.focused = validate_focused(focused)
        
        # name
        try:
            name = keyword_parameters.pop('name')
        except KeyError:
            pass
        else:
            self.name = validate_name(name)
        
        # options
        try:
            options = keyword_parameters.pop('options')
        except KeyError:
            pass
        else:
            self.options = validate_options(options)
        
        # type
        try:
            type_ = keyword_parameters.pop('type_')
        except KeyError:
            pass
        else:
            self.type = validate_type(type_)
        
        # value
        try:
            value = keyword_parameters.pop('value')
        except KeyError:
            pass
        else:
            self.value = validate_value(value)
        
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused keyword parameters: {keyword_parameters!r}.'
            )
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new interaction option from the data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`)
            Interaction option data.
        """
        self = object.__new__(cls)
        self.focused = parse_focused(data)
        self.name = parse_name(data)
        self.options = parse_options(data)
        self.type = parse_type(data)
        self.value = parse_value(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the interaction option into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        data = {}
        put_focused_into(self.focused, data, defaults)
        put_name_into(self.name, data, defaults)
        put_options_into(self.options, data, defaults)
        put_type_into(self.type, data, defaults)
        put_value_into(self.value, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the interaction option's representation."""
        repr_parts = [
            '<', self.__class__.__name__,
            ' name=', repr(self.name),
        ]
        
        if self.focused:
            repr_parts.append(' (focused)')
        
        value = self.value
        if (value is not None):
            repr_parts.append(', value=')
            repr_parts.append(reprlib.repr(value))
        
        type_ = self.type
        repr_parts.append(', type=')
        repr_parts.append(type_.name)
        repr_parts.append(' (')
        repr_parts.append(repr(type_.value))
        repr_parts.append(')')
        
        options = self.options
        if (options is not None):
            repr_parts.append(', options=[')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the interaction option's hash value."""
        hash_value = 0
        
        # focused
        hash_value ^= (self.focused << 16)
        
        # name
        hash_value ^= hash(self.name)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= (len(options) << 8)
            
            for option in options:
                hash_value ^= hash(option)
        
        # type
        hash_value ^= self.type.value
        
        # value
        value = self.value
        if (value is not None):
            hash_value ^= hash(value)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two interaction option's are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # focused
        if self.focused != other.focused:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # options
        if self.options != other.options:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # value
        if self.value != other.value:
            return False
        
        return True
    
    
    def iter_options(self):
        """
        Iterates over the sub-options of the interaction option.
        
        This method is an iterable generator.
        
        Yields
        ------
        option : ``InteractionOption``
        """
        options = self.options
        if options is not None:
            yield from options
    
    
    @property
    def focused_option(self):
        """
        Returns the focused option of the interaction option.
        
        Returns
        -------
        option : `None`, `instance<cls>`
        """
        if self.focused:
            return self
        
        for option in self.iter_options():
            focused_option =  option.focused_option
            if (focused_option is not None):
                return focused_option
    
    
    def _iter_non_focused_values(self):
        """
        Iterates over the non focused values of the interaction option.
        
        This method is an iterable generator.
        
        Yields
        ------
        name : `str`
            The option's name.
        value : `None`, `str`
            The option's value.
        """
        type_ = self.type
        if (
            (type_ is ApplicationCommandOptionType.sub_command) or
            (type_ is ApplicationCommandOptionType.sub_command_group)
        ):
            for option in self.iter_options():
                yield from option._iter_non_focused_values()
        else:
            if not self.focused:
                yield self.name, self.value
    
    
    def get_value_of(self, *option_names):
        """
        Gets the value for the option by the given name stack.
        
        Parameters
        ----------
        *option_names : `str`
            The option(s)'s name.
        
        Returns
        -------
        value : `None`, `str`
            The value, the user has been typed.
        """
        if option_names:
            option_name, *option_names = option_names
            
            for option in self.iter_options():
                if option.name == option_name:
                    value = option.get_value_of(*option_names)
                    break
            else:
                value = None
        else:
            value = self.value
        
        return value
